"""
multicast.py

Created by Thomas Morin on 2014-06-23.
Copyright (c) 2014-2017 Orange. All rights reserved.
License: 3-clause BSD. (See the COPYRIGHT file)
"""

from exabgp.protocol.ip import IP
from exabgp.util import character
from exabgp.util import ordinal
from exabgp.util import concat_bytes
from exabgp.bgp.message.update.nlri.qualifier import RouteDistinguisher
from exabgp.bgp.message.update.nlri.qualifier import EthernetTag

from exabgp.bgp.message.update.nlri.evpn.nlri import EVPN

# +---------------------------------------+
# |      RD   (8 octets)                  |
# +---------------------------------------+
# |  Ethernet Tag ID (4 octets)           |
# +---------------------------------------+
# |  IP Address Length (1 octet)          |
# +---------------------------------------+
# |   Originating Router's IP Addr        |
# |          (4 or 16 octets)             |
# +---------------------------------------+

# ===================================================================== EVPNNLRI


@EVPN.register
class Multicast(EVPN):
    CODE = 3
    NAME = "Inclusive Multicast Ethernet Tag"
    SHORT_NAME = "Multicast"

    def __init__(self, rd, etag, ip, packed=None, nexthop=None, action=None, addpath=None):
        EVPN.__init__(self, action, addpath)
        self.nexthop = nexthop
        self.rd = rd
        self.etag = etag
        self.ip = ip
        self._pack(packed)

    def __ne__(self, other):
        return not self.__eq__(other)

    def __str__(self):
        return "%s:%s:%s:%s" % (self._prefix(), self.rd._str(), self.etag, self.ip,)

    def __hash__(self):
        return hash((self.afi, self.safi, self.CODE, self.rd, self.etag, self.ip))

    def _pack(self, packed=None):
        if self._packed:
            return self._packed

        if packed:
            self._packed = packed
            return packed

        self._packed = concat_bytes(self.rd.pack(), self.etag.pack(), character(len(self.ip) * 8), self.ip.pack())
        return self._packed

    @classmethod
    def unpack(cls, data):
        rd = RouteDistinguisher.unpack(data[:8])
        etag = EthernetTag.unpack(data[8:12])
        iplen = ordinal(data[12])
        if iplen not in (4 * 8, 16 * 8):
            raise Exception("IP len is %d, but EVPN route currently support only IPv4" % iplen)
        ip = IP.unpack(data[13 : 13 + iplen // 8])
        return cls(rd, etag, ip, data)

    def json(self, compact=None):
        content = ' "code": %d, ' % self.CODE
        content += '"parsed": true, '
        content += '"raw": "%s", ' % self._raw()
        content += '"name": "%s", ' % self.NAME
        content += '%s, ' % self.rd.json()
        content += self.etag.json()
        if self.ip:
            content += ', "ip": "%s"' % str(self.ip)
        return '{%s }' % content



#!/usr/local/bin/python3.13

from __future__ import print_function

import sys
import socket

from exabgp.netlink import NetMask
from exabgp.netlink.attributes import Attributes

from exabgp.netlink.route.link import Link
from exabgp.netlink.route.neighbor import Neighbor
from exabgp.netlink.route.address import Address
from exabgp.netlink.route.network import Network


def usage():
    print('%s' % sys.argv[0])
    print('  addr  : show the ip address on the interface')
    print('  route : show the ip routing')


def addresses():
    links = {}
    for ifi in Link.getLinks():
        links[ifi.index] = ifi

    addrs = {}
    for ifa in Address.getAddresses():
        addrs.setdefault(ifa.index, []).append(ifa)

    neighbors = {}
    for neighbor in Neighbor.getNeighbors():
        neighbors.setdefault(neighbor.index, []).append(neighbor)

    for index, ifi in links.items():
        hwaddr = '<no addr>'
        if Address.Type.Attribute.IFLA_ADDRESS in ifi.attributes:
            hwaddr = ':'.join(x.encode('hex') for x in ifi.attributes[Address.Type.Attribute.IFLA_ADDRESS])
        print("%d: %s %s" % (ifi.index, ifi.attributes[Address.Type.Attribute.IFLA_IFNAME][:-1], hwaddr))

        for ifa in addrs.get(ifi.index, {}):
            address = ifa.attributes.get(Attributes.Type.IFA_ADDRESS)
            if not address:
                continue

            if ifa.family == socket.AF_INET:
                print('  %s %s' % ('inet ', socket.inet_ntop(ifa.family, address)))
            elif ifa.family == socket.AF_INET6:
                print('  %s %s' % ('inet6', socket.inet_ntop(ifa.family, address)))
            else:
                print('  %d %s' % (ifa.family, address.encode('hex')))

        for neighbor in neighbors.get(ifi.index, {}):
            if neighbor.state == Neighbor.Type.State.NUD_REACHABLE:
                address = neighbor.attributes.get(Neighbor.Type.Flag.NTF_USE, '\0\0\0\0')
                if ifa.family == socket.AF_INET:
                    print('  %s %s' % ('inet ', socket.inet_ntop(neighbor.family, address)), end=' ')
                elif ifa.family == socket.AF_INET6:
                    print('  %s %s' % ('inet ', socket.inet_ntop(neighbor.family, address)), end=' ')
                else:
                    print('  %d %s' % (ifa.family, address.encode('hex')))
                print('mac', ':'.join(_.encode('hex') for _ in neighbor.attributes[Neighbor.Type.State.NUD_REACHABLE]))


def routes():
    links = {}
    for ifi in Link.getLinks():
        links[ifi.index] = ifi.attributes.get(Link.Type.Attribute.IFLA_IFNAME).strip('\0')

    print('Kernel IP routing table')
    print('%-18s %-18s %-18s %-7s %s' % ('Destination', 'Genmask', 'Gateway', 'Metric', 'Iface'))

    for route in Network.getRoutes():
        if route.family != socket.AF_INET:
            continue

        if route.type not in (Network.Type.Type.RTN_LOCAL, Network.Type.Type.RTN_UNICAST):
            continue

        if route.src_len == 32:
            continue

        destination = route.attributes.get(Network.Type.Attribute.RTA_DST)
        gateway = route.attributes.get(Network.Type.Attribute.RTA_GATEWAY)

        oif = ord(route.attributes.get(Network.Type.Attribute.RTA_OIF)[0])
        metric = ord(route.attributes.get(Network.Type.Attribute.RTA_PRIORITY, '\0')[0])

        dst = '%s' % socket.inet_ntop(route.family, destination) if destination else ''
        gw = '%s' % socket.inet_ntop(route.family, gateway) if gateway else '0.0.0.0'
        mask = NetMask.CIDR[route.src_len]
        iface = links[oif]

        print('%-18s %-18s %-18s %-7d %-s' % (dst or '0.0.0.0', mask, gw, metric, iface))
        # if gateway: print route


def new():
    links = {}
    for ifi in Link.getLinks():
        links[ifi.index] = ifi.attributes.get(Link.Type.Attribute.IFLA_IFNAME).strip('\0')

    for route in Network.newRoute():
        print(route)

        # if route.family != socket.AF_INET:
        # 	continue
        #
        # if route.type not in (Network.Type.Type.RTN_LOCAL,Network.Type.Type.RTN_UNICAST):
        # 	continue
        #
        # if route.src_len == 32:
        # 	continue
        #
        # destination = route.attributes.get(Network.Type.Attribute.RTA_DST)
        # gateway = route.attributes.get(Network.Type.Attribute.RTA_GATEWAY)
        #
        # oif = ord(route.attributes.get(Network.Type.Attribute.RTA_OIF)[0])
        # metric = ord(route.attributes.get(Network.Type.Attribute.RTA_PRIORITY,'\0')[0])
        #
        # dst = '%s' % socket.inet_ntop(route.family, destination) if destination else ''
        # gw  = '%s' % socket.inet_ntop(route.family, gateway) if gateway else '0.0.0.0'
        # mask = NetMask.CIDR[route.src_len]
        # iface = links[oif]
        #
        # print '%-18s %-18s %-18s %-7d %-s' % (dst or '0.0.0.0',mask,gw,metric,iface)


def main():
    if len(sys.argv) < 2:
        usage()
        sys.exit(1)
    if 'addr'.startswith(sys.argv[1]):
        if len(sys.argv) == 2:
            addresses()
            sys.exit(0)
        if sys.argv[2] in 'show':
            addresses()
            sys.exit(0)
    if 'route'.startswith(sys.argv[1]):
        if len(sys.argv) == 2:
            routes()
            sys.exit(0)
        if 'show'.startswith(sys.argv[2]):
            routes()
            sys.exit(0)
        if 'add'.startswith(sys.argv[2]):
            new()
            sys.exit(0)
        if 'delete'.startswith(sys.argv[2]):
            print('adding')

    usage()
    sys.exit(0)


if __name__ == '__main__':
    main()



# encoding: utf-8
"""
bgp.py

Created by Thomas Mangin on 2009-08-30.
Copyright (c) 2009-2017 Exa Networks. All rights reserved.
License: 3-clause BSD. (See the COPYRIGHT file)
"""

from operator import eq
import os
import sys
import stat
import platform
import syslog
import string

from exabgp.util.dns import warn
from exabgp.logger import Logger

from exabgp.version import version

# import before the fork to improve copy on write memory savings
from exabgp.reactor.loop import Reactor

from exabgp.vendoring import docopt

from exabgp.configuration.usage import usage

from exabgp.debug import setup_report

setup_report()


def is_bgp(s):
    return all(c in string.hexdigits or c == ':' for c in s)


def __exit(memory, code):
    if memory:
        from exabgp.vendoring import objgraph

        sys.stdout.write('memory utilisation\n\n')
        sys.stdout.write(objgraph.show_most_common_types(limit=20))
        sys.stdout.write('\n\n\n')
        sys.stdout.write('generating memory utilisation graph\n\n')
        sys.stdout.write()
        obj = objgraph.by_type('Reactor')
        objgraph.show_backrefs([obj], max_depth=10)
    sys.exit(code)


def named_pipe(root, pipename='exabgp'):
    locations = [
        '/run/exabgp/',
        '/run/%d/' % os.getuid(),
        '/run/',
        '/var/run/exabgp/',
        '/var/run/%d/' % os.getuid(),
        '/var/run/',
        root + '/run/exabgp/',
        root + '/run/%d/' % os.getuid(),
        root + '/run/',
        root + '/var/run/exabgp/',
        root + '/var/run/%d/' % os.getuid(),
        root + '/var/run/',
    ]
    for location in locations:
        cli_in = location + pipename + '.in'
        cli_out = location + pipename + '.out'

        try:
            if not stat.S_ISFIFO(os.stat(cli_in).st_mode):
                continue
            if not stat.S_ISFIFO(os.stat(cli_out).st_mode):
                continue
        except KeyboardInterrupt:
            raise
        except Exception:
            continue
        os.environ['exabgp_cli_pipe'] = location
        return [location]
    return locations


def root_folder(options, locations):
    if options['--root']:
        return os.path.realpath(os.path.normpath(options['--root'])).rstrip('/')

    argv = os.path.realpath(os.path.normpath(os.path.join(os.getcwd(), sys.argv[0])))

    for location in locations:
        if argv.endswith(location):
            return argv[: -len(location)]
    return ''


def get_envfile(options, etc):
    envfile = 'exabgp.env' if not options["--env"] else options["--env"]
    if not envfile.startswith('/'):
        envfile = '%s/%s' % (etc, envfile)
    return envfile


def get_env(envfile):
    from exabgp.configuration.setup import environment

    try:
        return environment.setup(envfile)
    except environment.Error as exc:
        sys.stdout.write(usage)
        sys.stdout.flush()
        print('\nconfiguration issue,', str(exc))
        sys.exit(1)


def main():
    major = int(sys.version[0])
    minor = int(sys.version[2])

    if major <= 2 and minor < 5:
        sys.stdout.write('This program can not work (is not tested) with your python version (< 2.5)\n')
        sys.stdout.flush()
        sys.exit(1)

    cli_named_pipe = os.environ.get('exabgp_cli_pipe', '')
    if cli_named_pipe:
        from exabgp.application.control import main as control

        control(cli_named_pipe)
        sys.exit(0)

    options = docopt.docopt(usage, help=False)

    if options["--run"]:
        sys.argv = sys.argv[sys.argv.index('--run') + 1 :]
        if sys.argv[0] == 'healthcheck':
            from exabgp.application import run_healthcheck

            run_healthcheck()
        elif sys.argv[0] == 'cli':
            from exabgp.application import run_cli

            run_cli()
        else:
            sys.stdout.write(usage)
            sys.stdout.flush()
            sys.exit(0)
        return

    root = root_folder(
        options, ['/bin/exabgp', '/sbin/exabgp', '/lib/exabgp/application/bgp.py', '/lib/exabgp/application/control.py']
    )
    prefix = '' if root == '/usr' else root
    etc = '/etc/exabgp'

    os.environ['EXABGP_ETC'] = etc  # This is not most pretty

    if options["--version"]:
        sys.stdout.write('ExaBGP : %s\n' % version)
        sys.stdout.write('Python : %s\n' % sys.version.replace('\n', ' '))
        sys.stdout.write('Uname  : %s\n' % ' '.join(platform.uname()[:5]))
        sys.stdout.write('Root   : %s\n' % root)
        sys.stdout.flush()
        sys.exit(0)

    envfile = get_envfile(options, etc)
    env = get_env(envfile)

    # Must be done before setting the logger as it modify its behaviour
    if options["--debug"]:
        env.log.all = True
        env.log.level = syslog.LOG_DEBUG

    logger = Logger()

    from exabgp.configuration.setup import environment

    if options["--decode"]:
        decode = ''.join(options["--decode"]).replace(':', '').replace(' ', '')
        if not is_bgp(decode):
            sys.stdout.write(usage)
            sys.stdout.write('Environment values are:\n%s\n\n' % '\n'.join(' - %s' % _ for _ in environment.default()))
            sys.stdout.write('The BGP message must be an hexadecimal string.\n\n')
            sys.stdout.write('All colons or spaces are ignored, for example:\n\n')
            sys.stdout.write('  --decode 001E0200000007900F0003000101\n')
            sys.stdout.write('  --decode 001E:02:0000:0007:900F:0003:0001:01\n')
            sys.stdout.write('  --decode FFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFF001E0200000007900F0003000101\n')
            sys.stdout.write('  --decode FFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFF:001E:02:0000:0007:900F:0003:0001:01\n')
            sys.stdout.write('  --decode \'FFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFF 001E02 00000007900F0003000101\n\'')
            sys.stdout.flush()
            sys.exit(1)
    else:
        decode = ''

    if options["--signal"] and options["--signal"].isdigit():
        duration = options["--signal"]
        pid = os.fork()
        if pid:
            import time
            import signal

            try:
                time.sleep(int(duration))
                os.kill(pid, signal.SIGUSR1)
            except KeyboardInterrupt:
                pass
            try:
                pid, code = os.wait()
                sys.exit(code)
            except KeyboardInterrupt:
                try:
                    pid, code = os.wait()
                    sys.exit(code)
                except Exception:
                    sys.exit(0)

    if options["--help"]:
        sys.stdout.write(usage)
        sys.stdout.write('Environment values are:\n' + '\n'.join(' - %s' % _ for _ in environment.default()))
        sys.stdout.flush()
        sys.exit(0)

    if options["--decode"]:
        env.log.parser = True
        env.debug.route = decode
        env.tcp.bind = ''

    if options["--profile"]:
        env.profile.enable = True
        if options["--profile"].lower() in ['1', 'true']:
            env.profile.file = True
        elif options["--profile"].lower() in ['0', 'false']:
            env.profile.file = False
        else:
            env.profile.file = options["--profile"]

    if envfile and not os.path.isfile(envfile):
        comment = 'environment file missing\ngenerate it using "exabgp --fi > %s"' % envfile
    else:
        comment = ''

    if options["--full-ini"] or options["--fi"]:
        for line in environment.iter_ini():
            sys.stdout.write('%s\n' % line)
            sys.stdout.flush()
        sys.exit(0)

    if options["--full-env"] or options["--fe"]:
        print()
        for line in environment.iter_env():
            sys.stdout.write('%s\n' % line)
            sys.stdout.flush()
        sys.exit(0)

    if options["--diff-ini"] or options["--di"]:
        for line in environment.iter_ini(True):
            sys.stdout.write('%s\n' % line)
            sys.stdout.flush()
        sys.exit(0)

    if options["--diff-env"] or options["--de"]:
        for line in environment.iter_env(True):
            sys.stdout.write('%s\n' % line)
            sys.stdout.flush()
        sys.exit(0)

    if options["--once"]:
        env.tcp.once = True

    if options["--pdb"]:
        # The following may fail on old version of python (but is required for debug.py)
        os.environ['PDB'] = 'true'
        env.debug.pdb = True

    if options["--test"]:
        env.debug.selfcheck = True
        env.log.parser = True

    if options["--memory"]:
        env.debug.memory = True

    configurations = []
    # check the file only once that we have parsed all the command line options and allowed them to run
    if options["<configuration>"]:
        for f in options["<configuration>"]:
            # some users are using symlinks for atomic change of the configuration file
            # using mv may however be better practice :p
            normalised = os.path.realpath(os.path.normpath(f))
            target = os.path.realpath(normalised)
            if os.path.isfile(target):
                configurations.append(normalised)
                continue
            if f.startswith('etc/exabgp'):
                normalised = os.path.join(etc, f[11:])
                if os.path.isfile(normalised):
                    configurations.append(normalised)
                    continue

            logger.debug('one of the arguments passed as configuration is not a file (%s)' % f, 'configuration')
            sys.exit(1)

    else:
        sys.stdout.write(usage)
        sys.stdout.write('Environment values are:\n%s\n\n' % '\n'.join(' - %s' % _ for _ in environment.default()))
        sys.stdout.write('no configuration file provided')
        sys.stdout.flush()
        sys.exit(1)

    from exabgp.bgp.message.update.attribute import Attribute

    Attribute.caching = env.cache.attributes

    if env.debug.rotate or len(configurations) == 1:
        run(env, comment, configurations, root, options["--validate"])

    if not (env.log.destination in ('syslog', 'stdout', 'stderr') or env.log.destination.startswith('host:')):
        logger.error('can not log to files when running multiple configuration (as we fork)', 'configuration')
        sys.exit(1)

    try:
        # run each configuration in its own process
        pids = []
        for configuration in configurations:
            pid = os.fork()
            if pid == 0:
                run(env, comment, [configuration], root, options["--validate"], os.getpid())
            else:
                pids.append(pid)

        # If we get a ^C / SIGTERM, ignore just continue waiting for our child process
        import signal

        signal.signal(signal.SIGINT, signal.SIG_IGN)

        # wait for the forked processes
        for pid in pids:
            os.waitpid(pid, 0)
    except OSError as exc:
        logger.critical('can not fork, errno %d : %s' % (exc.errno, exc.strerror), 'reactor')
        sys.exit(1)


def run(env, comment, configurations, root, validate, pid=0):
    logger = Logger()

    logger.notice('Thank you for using ExaBGP', 'welcome')
    logger.notice('%s' % version, 'version')
    logger.notice('%s' % sys.version.replace('\n', ' '), 'interpreter')
    logger.notice('%s' % ' '.join(platform.uname()[:5]), 'os')
    logger.notice('%s' % root, 'installation')

    if comment:
        logger.notice(comment, 'advice')

    warning = warn()
    if warning:
        logger.warning(warning, 'advice')

    if env.api.cli:
        pipename = 'exabgp' if env.api.pipename is None else env.api.pipename
        pipes = named_pipe(root, pipename)
        if len(pipes) != 1:
            env.api.cli = False
            logger.error(
                'could not find the named pipes (%s.in and %s.out) required for the cli' % (pipename, pipename), 'cli'
            )
            logger.error('we scanned the following folders (the number is your PID):', 'cli')
            for location in pipes:
                logger.error(' - %s' % location, 'cli control')
            logger.error('please make them in one of the folder with the following commands:', 'cli control')
            logger.error('> mkfifo %s/run/%s.{in,out}' % (os.getcwd(), pipename), 'cli control')
            logger.error('> chmod 600 %s/run/%s.{in,out}' % (os.getcwd(), pipename), 'cli control')
            if os.getuid() != 0:
                logger.error(
                    '> chown %d:%d %s/run/%s.{in,out}' % (os.getuid(), os.getgid(), os.getcwd(), pipename),
                    'cli control',
                )
        else:
            pipe = pipes[0]
            os.environ['exabgp_cli_pipe'] = pipe
            os.environ['exabgp_api_pipename'] = pipename

            logger.info('named pipes for the cli are:', 'cli control')
            logger.info('to send commands  %s%s.in' % (pipe, pipename), 'cli control')
            logger.info('to read responses %s%s.out' % (pipe, pipename), 'cli control')

    if not env.profile.enable:
        exit_code = Reactor(configurations).run(validate, root)
        __exit(env.debug.memory, exit_code)

    try:
        import cProfile as profile
    except ImportError:
        import profile

    if env.profile.file == 'stdout':
        profiled = 'Reactor(%s).run(%s,"%s")' % (str(configurations), str(validate), str(root))
        exit_code = profile.run(profiled)
        __exit(env.debug.memory, exit_code)

    if pid:
        profile_name = "%s-pid-%d" % (env.profile.file, pid)
    else:
        profile_name = env.profile.file

    notice = ''
    if os.path.isdir(profile_name):
        notice = 'profile can not use this filename as output, it is not a directory (%s)' % profile_name
    if os.path.exists(profile_name):
        notice = 'profile can not use this filename as output, it already exists (%s)' % profile_name

    if not notice:
        cwd = os.getcwd()
        logger.debug('profiling ....', 'reactor')
        profiler = profile.Profile()
        profiler.enable()
        try:
            exit_code = Reactor(configurations).run(validate, root)
        except Exception:
            exit_code = Reactor.Exit.unknown
            raise
        finally:
            from exabgp.vendoring import lsprofcalltree

            profiler.disable()
            kprofile = lsprofcalltree.KCacheGrind(profiler)
            try:
                destination = profile_name if profile_name.startswith('/') else os.path.join(cwd, profile_name)
                with open(destination, 'w+') as write:
                    kprofile.output(write)
            except IOError:
                notice = 'could not save profiling in formation at: ' + destination
                logger.debug("-" * len(notice), 'reactor')
                logger.debug(notice, 'reactor')
                logger.debug("-" * len(notice), 'reactor')
            __exit(env.debug.memory, exit_code)
    else:
        logger.debug("-" * len(notice), 'reactor')
        logger.debug(notice, 'reactor')
        logger.debug("-" * len(notice), 'reactor')
        Reactor(configurations).run(validate, root)
        __exit(env.debug.memory, 1)


if __name__ == '__main__':
    main()



# encoding: utf-8
"""
igptags.py

Created by Evelio Vila on 2016-12-01.
Copyright (c) 2014-2017 Exa Networks. All rights reserved.
"""

from struct import unpack

from exabgp.bgp.message.notification import Notify
from exabgp.bgp.message.update.attribute.bgpls.linkstate import LINKSTATE

#   The IGP Route Tag TLV carries original IGP Tags (IS-IS [RFC5130] or
#   OSPF) of the prefix and is encoded as follows:
#
#      0                   1                   2                   3
#      0 1 2 3 4 5 6 7 8 9 0 1 2 3 4 5 6 7 8 9 0 1 2 3 4 5 6 7 8 9 0 1
#     +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
#     |              Type             |             Length            |
#     +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
#     //                    Route Tags (one or more)                 //
#     +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
#     https://tools.ietf.org/html/rfc7752#section-3.3.3.2


@LINKSTATE.register()
class IgpTags(object):
    TLV = 1153

    def __init__(self, igptags):
        self.igptags = igptags

    def __repr__(self):
        return "IGP Route Tags: %s" % (self.igptags)

    @classmethod
    def unpack(cls, data, length):
        tags = []
        n = length // 4
        ind = 0
        for i in list(range(n)):
            tag = unpack("!L", data[ind : 4 * (i + 1)])[0]
            tags.append(tag)
            ind += 4
        return cls(igptags=tags)

    def json(self, compact=None):
        return '"igp-route-tags": %s' % self.igptags



# encoding: utf-8
"""
traffic.py

Created by Thomas Mangin on 2014-06-21.
Copyright (c) 2014-2017 Exa Networks. All rights reserved.
License: 3-clause BSD. (See the COPYRIGHT file)
"""

import socket

from struct import pack
from struct import unpack

from exabgp.protocol.ip import IPv4
from exabgp.protocol.ip import IPv6
from exabgp.bgp.message.open.asn import ASN
from exabgp.bgp.message.open.capability.asn4 import ASN4
from exabgp.bgp.message.update.attribute.community.extended import ExtendedCommunity
from exabgp.bgp.message.update.attribute.community.extended import ExtendedCommunityIPv6


# ================================================================== TrafficRate
# RFC 5575


@ExtendedCommunity.register
class TrafficRate(ExtendedCommunity):
    COMMUNITY_TYPE = 0x80
    COMMUNITY_SUBTYPE = 0x06

    __slots__ = ['asn', 'rate']

    def __init__(self, asn, rate, community=None):
        self.asn = asn
        self.rate = rate
        ExtendedCommunity.__init__(
            self, community if community is not None else pack("!2sHf", self._subtype(), asn, rate)
        )

    def __repr__(self):
        return "rate-limit:%d" % self.rate

    @staticmethod
    def unpack(data):
        asn, rate = unpack('!Hf', data[2:8])
        return TrafficRate(ASN(asn), rate, data[:8])


# ================================================================ TrafficAction
# RFC 5575


@ExtendedCommunity.register
class TrafficAction(ExtendedCommunity):
    COMMUNITY_TYPE = 0x80
    COMMUNITY_SUBTYPE = 0x07

    _sample = {
        False: 0x0,
        True: 0x2,
    }

    _terminal = {
        False: 0x0,
        True: 0x1,
    }

    __slots__ = ['sample', 'terminal']

    def __init__(self, sample, terminal, community=None):
        self.sample = sample
        self.terminal = terminal
        bitmask = self._sample[sample] | self._terminal[terminal]
        ExtendedCommunity.__init__(
            self, community if community is not None else pack('!2sLBB', self._subtype(), 0, 0, bitmask)
        )

    def __repr__(self):
        s = []
        if self.sample:
            s.append('sample')
        if self.terminal:
            s.append('terminal')
        return 'action %s' % '-'.join(s)

    @staticmethod
    def unpack(data):
        (bit,) = unpack('!B', data[7:8])
        sample = bool(bit & 0x02)
        terminal = bool(bit & 0x01)
        return TrafficAction(sample, terminal, data[:8])


# ============================================================== TrafficRedirect
# RFC 5575 and 7674


@ExtendedCommunity.register
class TrafficRedirect(ExtendedCommunity):
    COMMUNITY_TYPE = 0x80
    COMMUNITY_SUBTYPE = 0x08

    __slots__ = ['asn', 'target']

    def __init__(self, asn, target, community=None):
        self.asn = asn
        self.target = target
        ExtendedCommunity.__init__(
            self, community if community is not None else pack("!2sHL", self._subtype(), asn, target)
        )

    def __repr__(self):
        return "redirect:%s:%s" % (self.asn, self.target)

    @staticmethod
    def unpack(data):
        asn, target = unpack('!HL', data[2:8])
        return TrafficRedirect(ASN(asn), target, data[:8])


@ExtendedCommunity.register
class TrafficRedirectASN4(ExtendedCommunity):
    COMMUNITY_TYPE = 0x82
    COMMUNITY_SUBTYPE = 0x08

    __slots__ = ['asn', 'target']

    def __init__(self, asn, target, community=None):
        self.asn = asn
        self.target = target
        ExtendedCommunity.__init__(
            self, community if community is not None else pack("!2sLH", self._subtype(), asn, target)
        )

    def __str__(self):
        return "redirect:%s:%s" % (self.asn, self.target)

    @staticmethod
    def unpack(data):
        asn, target = unpack('!LH', data[2:8])
        return TrafficRedirectASN4(ASN4(asn), target, data[:8])


# ================================================================== TrafficMark
# RFC 5575


@ExtendedCommunity.register
class TrafficMark(ExtendedCommunity):
    COMMUNITY_TYPE = 0x80
    COMMUNITY_SUBTYPE = 0x09

    __slots__ = ['dscp']

    def __init__(self, dscp, community=None):
        self.dscp = dscp
        ExtendedCommunity.__init__(
            self, community if community is not None else pack("!2sLBB", self._subtype(), 0, 0, dscp)
        )

    def __repr__(self):
        return "mark %d" % self.dscp

    @staticmethod
    def unpack(data):
        (dscp,) = unpack('!B', data[7:8])
        return TrafficMark(dscp, data[:8])


# =============================================================== TrafficNextHopIPv4IETF
# draft-ietf-idr-flowspec-redirect-02
# see RFC 4360 for ipv4 address specific extended community format


@ExtendedCommunity.register
class TrafficNextHopIPv4IETF(ExtendedCommunity):
    COMMUNITY_TYPE = 0x01
    COMMUNITY_SUBTYPE = 0x0C

    __slots__ = ['ip', 'copy']

    def __init__(self, ip, copy, community=None):
        self.ip = ip
        self.copy = copy
        ExtendedCommunity.__init__(
            self, community if community is not None else pack("!2s4sH", self._subtype(), ip.pack(), 1 if copy else 0)
        )

    def __repr__(self):
        return "copy-to-nexthop-ietf %s (with copy)" % self.ip if self.copy else "redirect-to-nexthop-ietf %s" % self.ip

    @staticmethod
    def unpack(data):
        ip, bit = unpack('!4sH', data[2:8])
        return TrafficNextHopIPv4IETF(IPv4.ntop(ip), bool(bit & 0x01), data[:8])


# =============================================================== TrafficNextHopIPv6IETF
# draft-ietf-idr-flowspec-redirect-02
# see RFC 5701 for ipv6 address specific extended community format


@ExtendedCommunityIPv6.register
class TrafficNextHopIPv6IETF(ExtendedCommunityIPv6):
    COMMUNITY_TYPE = 0x00
    COMMUNITY_SUBTYPE = 0x0C

    __slots__ = ['ip', 'copy']

    def __init__(self, ip, copy, community=None):
        self.ip = ip
        self.copy = copy
        ExtendedCommunityIPv6.__init__(
            self, community if community is not None else pack("!2s16sH", self._subtype(), ip.pack(), 1 if copy else 0)
        )

    def __repr__(self):
        return "copy-to-nexthop-ietf %s (with copy)" % self.ip if self.copy else "redirect-to-nexthop-ietf %s" % self.ip

    @staticmethod
    def unpack(data):
        ip, bit = unpack('!16sH', data[2:20])
        return TrafficNextHopIPv6IETF(IPv6.ntop(ip), bool(bit & 0x01), data[:20])


# =============================================================== TrafficNextHopSimpson
# draft-simpson-idr-flowspec-redirect-02

# XXX: FIXME: I guess this should be a subclass of NextHop or IP ..


@ExtendedCommunity.register
class TrafficNextHopSimpson(ExtendedCommunity):
    COMMUNITY_TYPE = 0x08
    COMMUNITY_SUBTYPE = 0x00

    __slots__ = ['copy']

    def __init__(self, copy, community=None):
        self.copy = copy
        ExtendedCommunity.__init__(
            self, community if community is not None else pack("!2sLH", self._subtype(), 0, 1 if copy else 0)
        )

    def __repr__(self):
        return "copy-to-nexthop" if self.copy else "redirect-to-nexthop"

    @staticmethod
    def unpack(data):
        (bit,) = unpack('!B', data[7:8])
        return TrafficNextHopSimpson(bool(bit & 0x01), data[:8])


# ============================================================ TrafficRedirectIPv6
# https://tools.ietf.org/html/rfc5701


@ExtendedCommunityIPv6.register
class TrafficRedirectIPv6(ExtendedCommunityIPv6):
    COMMUNITY_TYPE = 0x80
    COMMUNITY_SUBTYPE = 0x0B

    def __init__(self, ip, asn, community=None):
        self.ip = ip
        self.asn = asn
        ExtendedCommunityIPv6.__init__(
            self,
            community
            if community is not None
            else pack("!BB16sH", 0x00, 0x02, socket.inet_pton(socket.AF_INET6, ip), asn),
        )

    def __str__(self):
        return "redirect %s:%d" % (self.ip, self.asn)

    @staticmethod
    def unpack(data):
        ip, asn = unpack('!16sH', data[2:11])
        return TrafficRedirectIPv6(socket.inet_ntoa(socket.AF_INET6, ip), asn, data[:11])


# ============================================================ TrafficRedirectIP
# RFC 5575
# If we need to provide the <IP>:<ASN> form for the FlowSpec Redirect ...

# import socket
# Do not use socket, use IPv4.ntop or pton

# TrafficRedirectASN = TrafficRedirect

# class TrafficRedirectIP (ExtendedCommunity):
# 	COMMUNITY_TYPE = 0x80
# 	COMMUNITY_SUBTYPE = 0x08

# 	def __init__ (self, ip, target, community=None):
# 		self.ip = ip
# 		self.target = target
# 		ExtendedCommunity.__init__(self,community if community is not None else pack("!BB4sH",0x80,0x08,socket.inet_pton(socket.AF_INET,ip),target))

# 	def __str__ (self):
# 		return "redirect %s:%d" % (self.ip,self.target)

# 	@staticmethod
# 	def unpack (data):
# 		ip,target = unpack('!4sH',data[2:8])
# 		return TrafficRedirectIP(socket.inet_ntop(socket.AF_INET,ip),target,data[:8])



# encoding: utf-8
"""
linkid.py

Created by Evelio Vila on 2016-11-26. eveliovila@gmail.com
Copyright (c) 2009-2017 Exa Networks. All rights reserved.
License: 3-clause BSD. (See the COPYRIGHT file)
"""

from struct import pack
from struct import unpack

from exabgp.bgp.message.notification import Notify
from exabgp.util import ordinal

#       0                   1                   2                   3
#       0 1 2 3 4 5 6 7 8 9 0 1 2 3 4 5 6 7 8 9 0 1 2 3 4 5 6 7 8 9 0 1
#      +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
#      |                  Link Local Identifier                        |
#      +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
#      |                  Link Remote Identifier                       |
#      +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
#      https://tools.ietf.org/html/rfc5307 sec 1.1
# ================================================================== Link Local/Remote Identifiers


class LinkIdentifier(object):
    def __init__(self, local_id, remote_id, packed=None):
        self.local_id = local_id
        self.remote_id = remote_id
        self._packed = packed

    @classmethod
    def unpack(cls, data):
        local_id = unpack('!L', data[:4])[0]
        remote_id = unpack('!L', data[4:8])[0]
        return cls(local_id=local_id, remote_id=remote_id)

    def json(self):
        content = '"link-local-id": %s, ' % self.local_id + '"link-remote-id": %s' % self.remote_id
        return content

    def __eq__(self, other):
        return (self.local_id == other.local_id) and (self.remote_id == other.remote_id)

    def __neq__(self, other):
        return self.local_id != other.local_id

    def __lt__(self, other):
        raise RuntimeError('Not implemented')

    def __le__(self, other):
        raise RuntimeError('Not implemented')

    def __gt__(self, other):
        raise RuntimeError('Not implemented')

    def __ge__(self, other):
        raise RuntimeError('Not implemented')

    def __str__(self):
        return ':'.join('%02X' % ordinal(_) for _ in self._packed)

    def __repr__(self):
        return self.__str__()

    def __len__(self):
        if not self._packed:
            return 0
        return len(self._packed)

    def __hash__(self):
        return hash(str(self))

    def pack(self):
        return self._packed



# encoding: utf-8
"""
delay.py

Created by Thomas Mangin on 2009-08-25.
Copyright (c) 2017-2017 Exa Networks. All rights reserved.
"""

import time


# ======================================================================== Delay
# Exponential backup for outgoing connection


class Delay(object):
    def __init__(self):
        self._time = time.time()
        self._next = 0

    def reset(self):
        self._time = time.time()
        self._next = 0

    def increase(self):
        self._time = time.time() + self._next
        self._next = min(int(1 + self._next * 1.2), 60)

    def backoff(self):
        return self._time > time.time()



# encoding: utf-8
"""
labels.py

Created by Thomas Mangin on 2012-07-08.
Copyright (c) 2009-2017 Exa Networks. All rights reserved.
License: 3-clause BSD. (See the COPYRIGHT file)
"""

from struct import pack
from struct import unpack

from exabgp.util import character
from exabgp.util import concat_bytes_i


# ======================================================================= Labels
# RFC 3107


def opt_raw_label(label, format=' (%d)'):
    return format % label if label else ''


class Labels(object):
    MAX = pow(2, 20) - 1

    __slots__ = ['labels', 'packed', '_len', 'raw_labels']

    def __init__(self, labels, bos=True, raw_labels=None):
        self.labels = labels
        self.raw_labels = raw_labels
        packed = []
        if raw_labels:
            for label in raw_labels:
                packed.append(pack('!L', label)[1:])
            # fill self.labels as well, not for packing, but to allow
            # consistent string representations
            if not self.labels:
                self.labels = [x >> 4 for x in self.raw_labels]
        else:
            for label in labels:
                # shift to 20 bits of the label to be at the top of three bytes and then truncate.
                packed.append(pack('!L', label << 4)[1:])
            # Mark the bottom of stack with the bit
            if packed and bos:
                packed.pop()
                packed.append(pack('!L', (label << 4) | 1)[1:])
            self.raw_labels = [None for _ in self.labels]
        self.packed = concat_bytes_i(packed)
        self._len = len(self.packed)

    def __eq__(self, other):
        return self.labels == other.labels

    def __neq__(self, other):
        return self.labels != other.labels

    def __lt__(self, other):
        raise RuntimeError('comparing EthernetTag for ordering does not make sense')

    def __le__(self, other):
        raise RuntimeError('comparing EthernetTag for ordering does not make sense')

    def __gt__(self, other):
        raise RuntimeError('comparing EthernetTag for ordering does not make sense')

    def __ge__(self, other):
        raise RuntimeError('comparing EthernetTag for ordering does not make sense')

    def pack(self):
        return self.packed

    def __len__(self):
        return self._len

    def json(self):
        if len(self.labels) >= 1:
            return '"label": [ %s ]' % ', '.join(
                ["[%d%s]" % (l, opt_raw_label(r, ', %d')) for (l, r) in zip(self.labels, self.raw_labels)]
            )
        else:
            return ''

    def __str__(self):
        if len(self.labels) > 1:
            return ' label [ %s ]' % ' '.join(
                ["%d%s" % (l, opt_raw_label(r)) for (l, r) in zip(self.labels, self.raw_labels)]
            )
        elif len(self.labels) == 1:
            return ' label %d%s' % (self.labels[0], opt_raw_label(self.raw_labels[0]))
        else:
            return ''

    def __repr__(self):
        if len(self.labels) > 1:
            return '[ %s ]' % ','.join(["%d%s" % (l, opt_raw_label(r)) for (l, r) in zip(self.labels, self.raw_labels)])
        elif len(self.labels) == 1:
            return '%d%s' % (self.labels[0], opt_raw_label(self.raw_labels[0]))
        else:
            return '[ ]'

    @classmethod
    def unpack(cls, data):
        labels = []
        raw_labels = []
        while len(data):
            label = unpack('!L', character(0) + data[:3])[0]
            data = data[3:]
            labels.append(label >> 4)
            raw_labels.append(label)
            if label & 0x001:
                break
        return cls(labels, raw_labels=raw_labels)


Labels.NOLABEL = Labels([])



# http://teethgrinder.co.uk/perm.php?a=Python-memory-leak-detector
import gc
import inspect


def dump():
    # force collection
    print("\nCollecting GARBAGE:")
    gc.collect()
    # prove they have been collected
    print("\nCollecting GARBAGE:")
    gc.collect()

    print("\nGARBAGE OBJECTS:")
    for x in gc.garbage:
        s = str(x)
        if len(s) > 80:
            s = "%s..." % s[:80]

        print("::", s)
        print("		type:", type(x))
        print("   referrers:", len(gc.get_referrers(x)))
        try:
            print("	is class:", inspect.isclass(type(x)))
            print("	  module:", inspect.getmodule(x))

            lines, line_num = inspect.getsourcelines(type(x))
            print("	line num:", line_num)
            for l in lines:
                print("		line:", l.rstrip("\n"))
        except Exception:
            pass

        print()



# From https://git.geekli.st/daveyoon/khan-academy-lite/blob/14ab3615f66bcdd737909e7951a250da8a10646b/python-packages/memory_profiler.py
# LICENCE: MIT

"""Profile the memory usage of a Python program"""

__version__ = '0.26'

_CMD_USAGE = "/usr/local/bin/python3.13 -m memory_profiler script_file.py"

import time, sys, os, pdb
import warnings
import linecache
import inspect
import subprocess
from copy import copy

# TODO: provide alternative when multprocessing is not available
try:
    from multiprocessing import Process, Pipe
except ImportError:
    from multiprocessing.dummy import Process, Pipe


try:
    import psutil

    def _get_memory(pid):
        process = psutil.Process(pid)
        try:
            mem = float(process.get_memory_info()[0]) / (1024 ** 2)
        except psutil.AccessDenied:
            mem = -1
        return mem


except ImportError:

    warnings.warn("psutil module not found. memory_profiler will be slow")

    if os.name == 'posix':

        def _get_memory(pid):
            # ..
            # .. memory usage in MB ..
            # .. this should work on both Mac and Linux ..
            # .. subprocess.check_output appeared in 2.7, using Popen ..
            # .. for backwards compatibility ..
            out = subprocess.Popen(['ps', 'v', '-p', str(pid)], stdout=subprocess.PIPE).communicate()[0].split(b'\n')
            try:
                vsz_index = out[0].split().index(b'RSS')
                return float(out[1].split()[vsz_index]) / 1024
            except Exception:
                return -1

    else:
        raise NotImplementedError('The psutil module is required for non-unix ' 'platforms')


class Timer(Process):
    """
    Fetch memory consumption from over a time interval
    """

    def __init__(self, monitor_pid, interval, pipe, *args, **kw):
        self.monitor_pid = monitor_pid
        self.interval = interval
        self.pipe = pipe
        self.cont = True
        super(Timer, self).__init__(*args, **kw)

    def run(self):
        m = _get_memory(self.monitor_pid)
        timings = [m]
        self.pipe.send(0)  # we're ready
        while not self.pipe.poll(self.interval):
            m = _get_memory(self.monitor_pid)
            timings.append(m)
        self.pipe.send(timings)


def memory_usage(proc=-1, interval=0.1, timeout=None):
    """
    Return the memory usage of a process or piece of code

    Parameters
    ----------
    proc : {int, string, tuple, subprocess.Popen}, optional
        The process to monitor. Can be given by an integer/string
        representing a PID, by a Popen object or by a tuple
        representing a Python function. The tuple contains three
        values (f, args, kw) and specifies to run the function
        f(*args, **kw).
        Set to -1 (default) for current process.

    interval : float, optional
        Interval at which measurements are collected.

    timeout : float, optional
        Maximum amount of time (in seconds) to wait before returning.

    Returns
    -------
    mem_usage : list of floating-poing values
        memory usage, in MB. It's length is always < timeout / interval
    """
    ret = []

    if timeout is not None:
        max_iter = int(timeout / interval)
    elif isinstance(proc, int):
        # external process and no timeout
        max_iter = 1
    else:
        # for a Python function wait until it finishes
        max_iter = float('inf')

    if hasattr(proc, '__call__'):
        proc = (proc, (), {})
    if isinstance(proc, (list, tuple)):
        if len(proc) == 1:
            f, args, kw = (proc[0], (), {})
        elif len(proc) == 2:
            f, args, kw = (proc[0], proc[1], {})
        elif len(proc) == 3:
            f, args, kw = (proc[0], proc[1], proc[2])
        else:
            raise ValueError

        aspec = inspect.getargspec(f)
        n_args = len(aspec.args)
        if aspec.defaults is not None:
            n_args -= len(aspec.defaults)
        if n_args != len(args):
            raise ValueError('Function expects %s value(s) but %s where given' % (n_args, len(args)))

        child_conn, parent_conn = Pipe()  # this will store Timer's results
        p = Timer(os.getpid(), interval, child_conn)
        p.start()
        parent_conn.recv()  # wait until we start getting memory
        f(*args, **kw)
        parent_conn.send(0)  # finish timing
        ret = parent_conn.recv()
        p.join(5 * interval)
    elif isinstance(proc, subprocess.Popen):
        # external process, launched from Python
        while True:
            ret.append(_get_memory(proc.pid))
            time.sleep(interval)
            if timeout is not None:
                max_iter -= 1
                if max_iter == 0:
                    break
            if proc.poll() is not None:
                break
    else:
        # external process
        if proc == -1:
            proc = os.getpid()
        if max_iter == -1:
            max_iter = 1
        counter = 0
        while counter < max_iter:
            counter += 1
            ret.append(_get_memory(proc))
            time.sleep(interval)
    return ret


# ..
# .. utility functions for line-by-line ..


def _find_script(script_name):
    """ Find the script.

    If the input is not a file, then $PATH will be searched.
    """
    if os.path.isfile(script_name):
        return script_name
    path = os.getenv('PATH', os.defpath).split(os.pathsep)
    for folder in path:
        if folder == '':
            continue
        fn = os.path.join(folder, script_name)
        if os.path.isfile(fn):
            return fn

    sys.stderr.write('Could not find script {0}\n'.format(script_name))
    raise SystemExit(1)


class LineProfiler:
    """ A profiler that records the amount of memory for each line """

    def __init__(self, **kw):
        self.functions = list()
        self.code_map = {}
        self.enable_count = 0
        self.max_mem = kw.get('max_mem', None)

    def __call__(self, func):
        self.add_function(func)
        f = self.wrap_function(func)
        f.__module__ = func.__module__
        f.__name__ = func.__name__
        f.__doc__ = func.__doc__
        f.__dict__.update(getattr(func, '__dict__', {}))
        return f

    def add_function(self, func):
        """ Record line profiling information for the given Python function.
        """
        try:
            # func_code does not exist in Python3
            code = func.__code__
        except AttributeError:
            import warnings

            warnings.warn("Could not extract a code object for the object %r" % (func,))
            return
        if code not in self.code_map:
            self.code_map[code] = {}
            self.functions.append(func)

    def wrap_function(self, func):
        """ Wrap a function to profile it.
        """

        def f(*args, **kwds):
            self.enable_by_count()
            try:
                result = func(*args, **kwds)
            finally:
                self.disable_by_count()
            return result

        return f

    def run(self, cmd):
        """ Profile a single executable statment in the main namespace.
        """
        import __main__

        main_dict = __main__.__dict__
        return self.runctx(cmd, main_dict, main_dict)

    def runctx(self, cmd, globals, locals):
        """ Profile a single executable statement in the given namespaces.
        """
        self.enable_by_count()
        try:
            exec(cmd, globals, locals)
        finally:
            self.disable_by_count()
        return self

    def runcall(self, func, *args, **kw):
        """ Profile a single function call.
        """
        # XXX where is this used ? can be removed ?
        self.enable_by_count()
        try:
            return func(*args, **kw)
        finally:
            self.disable_by_count()

    def enable_by_count(self):
        """ Enable the profiler if it hasn't been enabled before.
        """
        if self.enable_count == 0:
            self.enable()
        self.enable_count += 1

    def disable_by_count(self):
        """ Disable the profiler if the number of disable requests matches the
        number of enable requests.
        """
        if self.enable_count > 0:
            self.enable_count -= 1
            if self.enable_count == 0:
                self.disable()

    def trace_memory_usage(self, frame, event, arg):
        """Callback for sys.settrace"""
        if event in ('line', 'return') and frame.f_code in self.code_map:
            lineno = frame.f_lineno
            if event == 'return':
                lineno += 1
            entry = self.code_map[frame.f_code].setdefault(lineno, [])
            entry.append(_get_memory(os.getpid()))

        return self.trace_memory_usage

    def trace_max_mem(self, frame, event, arg):
        # run into PDB as soon as memory is higher than MAX_MEM
        if event in ('line', 'return') and frame.f_code in self.code_map:
            c = _get_memory(os.getpid())
            if c >= self.max_mem:
                t = 'Current memory {0:.2f} MB exceeded the maximum '.format(c) + 'of {0:.2f} MB\n'.format(self.max_mem)
                sys.stdout.write(t)
                sys.stdout.write('Stepping into the debugger \n')
                frame.f_lineno -= 2
                p = pdb.Pdb()
                p.quitting = False
                p.stopframe = frame
                p.returnframe = None
                p.stoplineno = frame.f_lineno - 3
                p.botframe = None
                return p.trace_dispatch

        return self.trace_max_mem

    def __enter__(self):
        self.enable_by_count()

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.disable_by_count()

    def enable(self):
        if self.max_mem is not None:
            sys.settrace(self.trace_max_mem)
        else:
            sys.settrace(self.trace_memory_usage)

    def disable(self):
        self.last_time = {}
        sys.settrace(None)


def show_results(prof, stream=None, precision=3):
    if stream is None:
        stream = sys.stdout
    template = '{0:>6} {1:>12} {2:>12}   {3:<}'

    for code in prof.code_map:
        lines = prof.code_map[code]
        if not lines:
            # .. measurements are empty ..
            continue
        filename = code.co_filename
        if filename.endswith((".pyc", ".pyo")):
            filename = filename[:-1]
        stream.write('Filename: ' + filename + '\n\n')
        if not os.path.exists(filename):
            stream.write('ERROR: Could not find file ' + filename + '\n')
            if filename.startswith("ipython-input") or filename.startswith("<ipython-input"):
                print(
                    "NOTE: %mprun can only be used on functions defined in "
                    "physical files, and not in the IPython environment."
                )
            continue
        all_lines = linecache.getlines(filename)
        sub_lines = inspect.getblock(all_lines[code.co_firstlineno - 1 :])
        linenos = range(code.co_firstlineno, code.co_firstlineno + len(sub_lines))
        lines_normalized = {}

        header = template.format('Line #', 'Mem usage', 'Increment', 'Line Contents')
        stream.write(header + '\n')
        stream.write('=' * len(header) + '\n')
        # move everything one frame up
        keys = sorted(lines.keys())

        k_old = keys[0] - 1
        lines_normalized[keys[0] - 1] = lines[keys[0]]
        for i in range(1, len(lines_normalized[keys[0] - 1])):
            lines_normalized[keys[0] - 1][i] = -1.0
        k = keys.pop(0)
        while keys:
            lines_normalized[k] = lines[keys[0]]
            for i in range(len(lines_normalized[k_old]), len(lines_normalized[k])):
                lines_normalized[k][i] = -1.0
            k_old = k
            k = keys.pop(0)

        first_line = sorted(lines_normalized.keys())[0]
        mem_old = max(lines_normalized[first_line])
        precision = int(precision)
        template_mem = '{{0:{0}.{1}'.format(precision + 6, precision) + 'f} MB'
        for i, l in enumerate(linenos):
            mem = ''
            inc = ''
            if l in lines_normalized:
                mem = max(lines_normalized[l])
                inc = mem - mem_old
                mem_old = mem
                mem = template_mem.format(mem)
                inc = template_mem.format(inc)
            stream.write(template.format(l, mem, inc, sub_lines[i]))
        stream.write('\n\n')


# A lprun-style %mprun magic for IPython.
def magic_mprun(self, parameter_s=''):
    """ Execute a statement under the line-by-line memory profiler from the
    memory_profilser module.

    Usage:
      %mprun -f func1 -f func2 <statement>

    The given statement (which doesn't require quote marks) is run via the
    LineProfiler. Profiling is enabled for the functions specified by the -f
    options. The statistics will be shown side-by-side with the code through
    the pager once the statement has completed.

    Options:

    -f <function>: LineProfiler only profiles functions and methods it is told
    to profile.  This option tells the profiler about these functions. Multiple
    -f options may be used. The argument may be any expression that gives
    a Python function or method object. However, one must be careful to avoid
    spaces that may confuse the option parser. Additionally, functions defined
    in the interpreter at the In[] prompt or via %run currently cannot be
    displayed.  Write these functions out to a separate file and import them.

    One or more -f options are required to get any useful results.

    -T <filename>: dump the text-formatted statistics with the code
    side-by-side out to a text file.

    -r: return the LineProfiler object after it has completed profiling.
    """
    try:
        from StringIO import StringIO
    except ImportError:  # Python 3.x
        from io import StringIO

    # Local imports to avoid hard dependency.
    from distutils.version import LooseVersion
    import IPython

    ipython_version = LooseVersion(IPython.__version__)
    if ipython_version < '0.11':
        from IPython.genutils import page
        from IPython.ipstruct import Struct
        from IPython.ipapi import UsageError
    else:
        from IPython.core.page import page
        from IPython.utils.ipstruct import Struct
        from IPython.core.error import UsageError

    # Escape quote markers.
    opts_def = Struct(T=[''], f=[])
    parameter_s = parameter_s.replace('"', r'\"').replace("'", r"\'")
    opts, arg_str = self.parse_options(parameter_s, 'rf:T:', list_all=True)
    opts.merge(opts_def)
    global_ns = self.shell.user_global_ns
    local_ns = self.shell.user_ns

    # Get the requested functions.
    funcs = []
    for name in opts.f:
        try:
            funcs.append(eval(name, global_ns, local_ns))
        except Exception as e:
            raise UsageError('Could not find function %r.\n%s: %s' % (name, e.__class__.__name__, e))

    profile = LineProfiler()
    for func in funcs:
        profile(func)

    # Add the profiler to the builtins for @profile.
    try:
        import builtins
    except ImportError:  # Python 3x
        import __builtin__ as builtins

    if 'profile' in builtins.__dict__:
        had_profile = True
        old_profile = builtins.__dict__['profile']
    else:
        had_profile = False
        old_profile = None
    builtins.__dict__['profile'] = profile

    try:
        try:
            profile.runctx(arg_str, global_ns, local_ns)
            message = ''
        except SystemExit:
            message = "*** SystemExit exception caught in code being profiled."
        except KeyboardInterrupt:
            message = "*** KeyboardInterrupt exception caught in code being " "profiled."
    finally:
        if had_profile:
            builtins.__dict__['profile'] = old_profile

    # Trap text output.
    stdout_trap = StringIO()
    show_results(profile, stdout_trap)
    output = stdout_trap.getvalue()
    output = output.rstrip()

    if ipython_version < '0.11':
        page(output, screen_lines=self.shell.rc.screen_length)
    else:
        page(output)
    print(message,)

    text_file = opts.T[0]
    if text_file:
        with open(text_file, 'w') as pfile:
            pfile.write(output)
        print('\n*** Profile printout saved to text file %s. %s' % (text_file, message))

    return_value = None
    if 'r' in opts:
        return_value = profile

    return return_value


def _func_exec(stmt, ns):
    # helper for magic_memit, just a function proxy for the exec
    # statement
    exec(stmt, ns)


# a timeit-style %memit magic for IPython
def magic_memit(self, line=''):
    """Measure memory usage of a Python statement

    Usage, in line mode:
      %memit [-r<R>t<T>] statement

    Options:
    -r<R>: repeat the loop iteration <R> times and take the best result.
    Default: 1

    -t<T>: timeout after <T> seconds. Default: None

    Examples
    --------
    ::

      In [1]: import numpy as np

      In [2]: %memit np.zeros(1e7)
      maximum of 1: 76.402344 MB per loop

      In [3]: %memit np.ones(1e6)
      maximum of 1: 7.820312 MB per loop

      In [4]: %memit -r 10 np.empty(1e8)
      maximum of 10: 0.101562 MB per loop

    """
    opts, stmt = self.parse_options(line, 'r:t', posix=False, strict=False)
    repeat = int(getattr(opts, 'r', 1))
    if repeat < 1:
        repeat = 1
    timeout = int(getattr(opts, 't', 0))
    if timeout <= 0:
        timeout = None

    mem_usage = []
    for _ in range(repeat):
        tmp = memory_usage((_func_exec, (stmt, self.shell.user_ns)), timeout=timeout)
        mem_usage.extend(tmp)

    if mem_usage:
        print('maximum of %d: %f MB per loop' % (repeat, max(mem_usage)))
    else:
        print('ERROR: could not read memory usage, try with a lower interval or more iterations')


def load_ipython_extension(ip):
    """This is called to load the module as an IPython extension."""
    ip.define_magic('mprun', magic_mprun)
    ip.define_magic('memit', magic_memit)


def profile(func, stream=None):
    """
    Decorator that will run the function and print a line-by-line profile
    """

    def wrapper(*args, **kwargs):
        prof = LineProfiler()
        val = prof(func)(*args, **kwargs)
        show_results(prof, stream=stream)
        return val

    return wrapper


if __name__ == '__main__':
    from optparse import OptionParser

    parser = OptionParser(usage=_CMD_USAGE, version=__version__)
    parser.disable_interspersed_args()
    parser.add_option(
        "--pdb-mmem",
        dest="max_mem",
        metavar="MAXMEM",
        type="float",
        action="store",
        help="step into the debugger when memory exceeds MAXMEM",
    )
    parser.add_option(
        '--precision',
        dest="precision",
        type="int",
        action="store",
        default=3,
        help="precision of memory output in number of significant digits",
    )

    if not sys.argv[1:]:
        parser.print_help()
        sys.exit(2)

    (options, args) = parser.parse_args()
    del sys.argv[0]  # Hide "memory_profiler.py" from argument list

    prof = LineProfiler(max_mem=options.max_mem)
    __file__ = _find_script(args[0])
    try:
        if sys.version_info[0] < 3:
            import __builtin__

            __builtin__.__dict__['profile'] = prof
            ns = copy(locals())
            ns['profile'] = prof  # shadow the profile decorator defined above
            execfile(__file__, ns, ns)
        else:
            import builtins

            builtins.__dict__['profile'] = prof
            ns = copy(locals())
            ns['profile'] = prof  # shadow the profile decorator defined above
            exec(compile(open(__file__).read(), __file__, 'exec'), ns, copy(globals()))
    finally:
        show_results(prof, precision=options.precision)



# encoding: utf-8
"""
announce/flow.py

Created by Thomas Mangin on 2017-07-06.
Copyright (c) 2009-2017 Exa Networks. All rights reserved.
License: 3-clause BSD. (See the COPYRIGHT file)
"""

from exabgp.rib.change import Change

from exabgp.bgp.message import OUT

from exabgp.protocol.family import AFI
from exabgp.protocol.family import SAFI

from exabgp.bgp.message.update.nlri.flow import Flow
from exabgp.bgp.message.update.attribute import Attributes

from exabgp.configuration.flow.parser import source
from exabgp.configuration.flow.parser import destination
from exabgp.configuration.flow.parser import any_port
from exabgp.configuration.flow.parser import source_port
from exabgp.configuration.flow.parser import destination_port
from exabgp.configuration.flow.parser import tcp_flags
from exabgp.configuration.flow.parser import protocol
from exabgp.configuration.flow.parser import next_header
from exabgp.configuration.flow.parser import fragment
from exabgp.configuration.flow.parser import packet_length
from exabgp.configuration.flow.parser import icmp_code
from exabgp.configuration.flow.parser import icmp_type
from exabgp.configuration.flow.parser import dscp
from exabgp.configuration.flow.parser import traffic_class
from exabgp.configuration.flow.parser import flow_label

from exabgp.configuration.flow.parser import accept
from exabgp.configuration.flow.parser import discard
from exabgp.configuration.flow.parser import rate_limit
from exabgp.configuration.flow.parser import redirect
from exabgp.configuration.flow.parser import redirect_next_hop
from exabgp.configuration.flow.parser import redirect_next_hop_ietf
from exabgp.configuration.flow.parser import copy
from exabgp.configuration.flow.parser import mark
from exabgp.configuration.flow.parser import action

from exabgp.configuration.static.parser import community
from exabgp.configuration.static.parser import large_community
from exabgp.configuration.static.parser import extended_community

from exabgp.configuration.flow.parser import interface_set

from exabgp.configuration.announce import ParseAnnounce


class AnnounceFlow(ParseAnnounce):
    definition = [
        'source 10.0.0.0/24',
        'source ::1/128/0',
        'destination 10.0.1.0/24',
        'port 25',
        'source-port >1024',
        'destination-port [ =80 =3128 >8080&<8088 ]',
        'packet-length [ >200&<300 >400&<500 ]',
        'tcp-flags [ 0x20+0x8+0x1 #name-here ]  # to check',
        '(ipv4 only) protocol [ udp tcp ]',
        '(ipv4 only) fragment [ dont-fragment is-fragment first-fragment last-fragment ]',
        '(ipv6 only) next-header [ udp tcp ]',
        '(ipv6 only) flow-label >100&<2000',
        '(ipv6 only) icmp-type 35  # to check',
        '(ipv6 only) icmp-code 55  # to check',
        'accept',
        'discard',
        'rate-limit 9600',
        'redirect 30740:12345',
        'redirect 1.2.3.4:5678',
        'redirect 1.2.3.4',
        'redirect-next-hop',
        'copy 1.2.3.4',
        'mark 123',
        'action sample|terminal|sample-terminal',
    ]

    syntax = 'flow {\n' '  <safi> %s;\n' '}' % ';\n  '.join(definition)

    known = {
        'source': source,
        'source-ipv4': source,
        'source-ipv6': source,
        'destination': destination,
        'destination-ipv4': destination,
        'destination-ipv6': destination,
        'protocol': protocol,
        'next-header': next_header,
        'port': any_port,
        'destination-port': destination_port,
        'source-port': source_port,
        'icmp-type': icmp_type,
        'icmp-code': icmp_code,
        'tcp-flags': tcp_flags,
        'packet-length': packet_length,
        'dscp': dscp,
        'traffic-class': traffic_class,
        'fragment': fragment,
        'flow-label': flow_label,
        'accept': accept,
        'discard': discard,
        'rate-limit': rate_limit,
        'redirect': redirect,
        'redirect-to-nexthop': redirect_next_hop,
        'redirect-to-nexthop-ietf': redirect_next_hop_ietf,
        'copy': copy,
        'mark': mark,
        'action': action,
        'community': community,
        'large-community': large_community,
        'extended-community': extended_community,
        'interface-set': interface_set,
    }

    # 'source-ipv4','destination-ipv4',

    action = {
        'source': 'nlri-add',
        'source-ipv4': 'nlri-add',
        'source-ipv6': 'nlri-add',
        'destination': 'nlri-add',
        'destination-ipv4': 'nlri-add',
        'destination-ipv6': 'nlri-add',
        'port': 'nlri-add',
        'source-port': 'nlri-add',
        'destination-port': 'nlri-add',
        'protocol': 'nlri-add',
        'packet-length': 'nlri-add',
        'tcp-flags': 'nlri-add',
        'next-header': 'nlri-add',
        'fragment': 'nlri-add',
        'icmp-code': 'nlri-add',
        'icmp-type': 'nlri-add',
        'packet-length': 'nlri-add',
        'dscp': 'nlri-add',
        'traffic-class': 'nlri-add',
        'flow-label': 'nlri-add',
        'accept': 'nop',
        'discard': 'attribute-add',
        'rate-limit': 'attribute-add',
        'redirect': 'nexthop-and-attribute',
        'redirect-to-nexthop': 'attribute-add',
        'redirect-to-nexthop-ietf': 'attribute-add',
        'copy': 'nexthop-and-attribute',
        'mark': 'attribute-add',
        'action': 'attribute-add',
        'community': 'attribute-add',
        'large-community': 'attribute-add',
        'extended-community': 'attribute-add',
        'interface-set': 'attribute-add',
    }

    assign = dict()
    default = dict()

    name = 'flow'

    def __init__(self, tokeniser, scope, error, logger):
        ParseAnnounce.__init__(self, tokeniser, scope, error, logger)

    def clear(self):
        pass

    def pre(self):
        self.scope.to_context(self.name)
        return True

    def post(self):
        self.scope.to_context(self.name)
        self.scope.set('routes', self.scope.pop('route', {}).get('routes', []))
        self.scope.extend('routes', self.scope.pop('flow', []))
        return True

    def check(self):
        return True


def flow(tokeniser, afi, safi):
    change = Change(Flow(afi, safi, OUT.ANNOUNCE), Attributes())

    while True:
        command = tokeniser()

        if not command:
            break

        action = AnnounceFlow.action[command]

        if action == 'nlri-add':
            for adding in AnnounceFlow.known[command](tokeniser):
                change.nlri.add(adding)
        elif action == 'attribute-add':
            change.attributes.add(AnnounceFlow.known[command](tokeniser))
        elif action == 'nexthop-and-attribute':
            nexthop, attribute = AnnounceFlow.known[command](tokeniser)
            change.nlri.nexthop = nexthop
            change.attributes.add(attribute)
        elif action == 'nop':
            pass  # yes nothing to do !
        else:
            raise ValueError('flow: unknown command "%s"' % command)

    return [change]


@ParseAnnounce.register('flow', 'extend-name', 'ipv4')
def flow_ip_v4(tokeniser):
    return flow(tokeniser, AFI.ipv4, SAFI.flow_ip)


@ParseAnnounce.register('flow-vpn', 'extend-name', 'ipv4')
def flow_vpn_v4(tokeniser):
    return flow(tokeniser, AFI.ipv4, SAFI.flow_vpn)


@ParseAnnounce.register('flow', 'extend-name', 'ipv6')
def flow_ip_v6(tokeniser):
    return flow(tokeniser, AFI.ipv6, SAFI.flow_ip)


@ParseAnnounce.register('flow-vpn', 'extend-name', 'ipv6')
def flow_vpn_v6(tokeniser):
    return flow(tokeniser, AFI.ipv6, SAFI.flow_vpn)



# encoding: utf-8
"""
setup.py

Created by Thomas Mangin on 2014-12-23.
Copyright (c) 2014-2017 Exa Networks. All rights reserved.
License: 3-clause BSD. (See the COPYRIGHT file)
"""

from exabgp.configuration.environment import environment

_SPACE = {'space': ' ' * 33}

HELP_STDOUT = (
    """\
where logging should log
%(space)s syslog (or no setting) sends the data to the local syslog syslog
%(space)s host:<location> sends the data to a remote syslog server
%(space)s stdout sends the data to stdout
%(space)s stderr sends the data to stderr
%(space)s <filename> send the data to a file"""
    % _SPACE
)


environment.application = 'exabgp'
environment.configuration = {
    'profile': {
        'enable': {
            'read': environment.boolean,
            'write': environment.lower,
            'value': 'false',
            'help': 'toggle profiling of the code',
        },
        'file': {
            'read': environment.unquote,
            'write': environment.quote,
            'value': '',
            'help': 'profiling result file, none means stdout, no overwriting',
        },
    },
    'pdb': {
        'enable': {
            'read': environment.boolean,
            'write': environment.lower,
            'value': 'false',
            'help': 'on program fault, start pdb the python interactive debugger',
        }
    },
    'daemon': {
        'pid': {
            'read': environment.unquote,
            'write': environment.quote,
            'value': '',
            'help': 'where to save the pid if we manage it',
        },
        'user': {'read': environment.user, 'write': environment.quote, 'value': 'nobody', 'help': 'user to run as',},
        'daemonize': {
            'read': environment.boolean,
            'write': environment.lower,
            'value': 'false',
            'help': 'should we run in the background',
        },
        'drop': {
            'read': environment.boolean,
            'write': environment.lower,
            'value': 'true',
            'help': 'drop privileges before forking processes',
        },
        'umask': {
            'read': environment.umask_read,
            'write': environment.umask_write,
            'value': '0137',
            'help': 'run daemon with this umask, governs perms of logfiles etc.',
        },
    },
    'log': {
        'enable': {'read': environment.boolean, 'write': environment.lower, 'value': 'true', 'help': 'enable logging',},
        'level': {
            'read': environment.syslog_value,
            'write': environment.syslog_name,
            'value': 'INFO',
            'help': 'log message with at least the priority SYSLOG.<level>',
        },
        'destination': {
            'read': environment.unquote,
            'write': environment.quote,
            'value': 'stdout',
            'help': HELP_STDOUT,
        },
        'all': {
            'read': environment.boolean,
            'write': environment.lower,
            'value': 'false',
            'help': 'report debug information for everything',
        },
        'configuration': {
            'read': environment.boolean,
            'write': environment.lower,
            'value': 'true',
            'help': 'report command parsing',
        },
        'reactor': {
            'read': environment.boolean,
            'write': environment.lower,
            'value': 'true',
            'help': 'report signal received, command reload',
        },
        'daemon': {
            'read': environment.boolean,
            'write': environment.lower,
            'value': 'true',
            'help': 'report pid change, forking, ...',
        },
        'processes': {
            'read': environment.boolean,
            'write': environment.lower,
            'value': 'true',
            'help': 'report handling of forked processes',
        },
        'network': {
            'read': environment.boolean,
            'write': environment.lower,
            'value': 'true',
            'help': 'report networking information (TCP/IP, network state,...)',
        },
        'packets': {
            'read': environment.boolean,
            'write': environment.lower,
            'value': 'false',
            'help': 'report BGP packets sent and received',
        },
        'rib': {
            'read': environment.boolean,
            'write': environment.lower,
            'value': 'false',
            'help': 'report change in locally configured routes',
        },
        'message': {
            'read': environment.boolean,
            'write': environment.lower,
            'value': 'false',
            'help': 'report changes in route announcement on config reload',
        },
        'timers': {
            'read': environment.boolean,
            'write': environment.lower,
            'value': 'false',
            'help': 'report keepalives timers',
        },
        'routes': {
            'read': environment.boolean,
            'write': environment.lower,
            'value': 'false',
            'help': 'report received routes',
        },
        'parser': {
            'read': environment.boolean,
            'write': environment.lower,
            'value': 'false',
            'help': 'report BGP message parsing details',
        },
        'short': {
            'read': environment.boolean,
            'write': environment.lower,
            'value': 'false',
            'help': 'use short log format (not prepended with time,level,pid and source)',
        },
    },
    'tcp': {
        'once': {
            'read': environment.boolean,
            'write': environment.lower,
            'value': 'false',
            'help': 'only one tcp connection attempt per peer (for debuging scripts)',
        },
        'delay': {
            'read': environment.integer,
            'write': environment.nop,
            'value': '0',
            'help': 'start to announce route when the minutes in the hours is a modulo of this number',
        },
        'bind': {
            'read': environment.ip_list,
            'write': environment.quote_list,
            'value': '',
            'help': 'Space separated list of IPs to bind on when listening (no ip to disable)',
        },
        'port': {
            'read': environment.integer,
            'write': environment.nop,
            'value': '179',
            'help': 'port to bind on when listening',
        },
        'acl': {
            'read': environment.boolean,
            'write': environment.lower,
            'value': '',
            'help': '(experimental) unimplemented',
        },
    },
    'bgp': {
        'openwait': {
            'read': environment.integer,
            'write': environment.nop,
            'value': '60',
            'help': 'how many seconds we wait for an open once the TCP session is established',
        },
    },
    'cache': {
        'attributes': {
            'read': environment.boolean,
            'write': environment.lower,
            'value': 'true',
            'help': 'cache all attributes (configuration and wire) for faster parsing',
        },
        'nexthops': {
            'read': environment.boolean,
            'write': environment.lower,
            'value': 'true',
            'help': 'cache routes next-hops (deprecated: next-hops are always cached)',
        },
    },
    'api': {
        'ack': {
            'read': environment.boolean,
            'write': environment.lower,
            'value': 'true',
            'help': 'acknowledge api command(s) and report issues',
        },
        'chunk': {
            'read': environment.integer,
            'write': environment.nop,
            'value': '1',
            'help': 'maximum lines to print before yielding in show routes api',
        },
        'encoder': {
            'read': environment.api,
            'write': environment.lower,
            'value': 'json',
            'help': '(experimental) default encoder to use with with external API (text or json)',
        },
        'compact': {
            'read': environment.boolean,
            'write': environment.lower,
            'value': 'false',
            'help': 'shorter JSON encoding for IPv4/IPv6 Unicast NLRI',
        },
        'respawn': {
            'read': environment.boolean,
            'write': environment.lower,
            'value': 'true',
            'help': 'should we try to respawn helper processes if they dies',
        },
        'terminate': {
            'read': environment.boolean,
            'write': environment.lower,
            'value': 'false',
            'help': 'should we terminate ExaBGP if any helper process dies',
        },
        'cli': {
            'read': environment.boolean,
            'write': environment.lower,
            'value': 'true',
            'help': 'should we create a named pipe for the cli',
        },
        'pipename': {
            'read': environment.unquote,
            'write': environment.quote,
            'value': 'exabgp',
            'help': 'name to be used for the exabgp pipe',
        },
    },
    'reactor': {
        'speed': {
            'read': environment.real,
            'write': environment.nop,
            'value': '1.0',
            'help': 'reactor loop time\n%(space)s use only if you understand the code.' % _SPACE,
        },
    },
    # Here for internal use
    'debug': {
        'pdb': {
            'read': environment.boolean,
            'write': environment.lower,
            'value': 'false',
            'help': 'enable python debugger on errors',
        },
        'memory': {
            'read': environment.boolean,
            'write': environment.lower,
            'value': 'false',
            'help': 'command line option --memory',
        },
        'configuration': {
            'read': environment.boolean,
            'write': environment.lower,
            'value': 'false',
            'help': 'undocumented option: raise when parsing configuration errors',
        },
        'selfcheck': {
            'read': environment.boolean,
            'write': environment.lower,
            'value': 'false',
            'help': 'does a self check on the configuration file',
        },
        'route': {
            'read': environment.unquote,
            'write': environment.quote,
            'value': '',
            'help': 'decode the route using the configuration',
        },
        'defensive': {
            'read': environment.boolean,
            'write': environment.lower,
            'value': 'false',
            'help': 'generate random fault in the code in purpose',
        },
        'rotate': {
            'read': environment.boolean,
            'write': environment.lower,
            'value': 'false',
            'help': 'rotate configurations file on reload (signal)',
        },
    },
}



# encoding: utf-8
"""
address.py

Created by Thomas Mangin on 2015-03-31.
Copyright (c) 2009-2017 Exa Networks. All rights reserved.
License: 3-clause BSD. (See the COPYRIGHT file)
"""

import socket
from struct import calcsize
from collections import namedtuple

from exabgp.netlink.message import Message


# 0                   1                   2                   3
# 0 1 2 3 4 5 6 7 8 9 0 1 2 3 4 5 6 7 8 9 0 1 2 3 4 5 6 7 8 9 0 1
# +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
# |   Family    |     Length    |     Flags     |    Scope      |
# +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
# |                     Interface Index                         |
# +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+


class Address(Message):
    class Header(object):
        PACK = '4Bi'
        LEN = calcsize(PACK)

    format = namedtuple('Address', 'family prefixlen flags scope index attributes')

    class Command(object):
        RTM_NEWADDR = 0x14
        RTM_DELADDR = 0x15
        RTM_GETADDR = 0x16

    class Type(object):
        class Family(object):
            AF_INET = socket.AF_INET
            AF_INET6 = socket.AF_INET6

        class Flag(object):
            IFA_F_SECONDARY = 0x00  # For secondary address (alias interface)
            IFA_F_PERMANENT = 0x00  # For a permanent address set by the user.  When this is not set, it means the address was dynamically created (e.g., by stateless autoconfiguration).
            IFA_F_DEPRECATED = 0x00  # Defines deprecated (IPV4) address
            IFA_F_TENTATIVE = (
                0x00  # Defines tentative (IPV4) address (duplicate address detection is still in progress)
            )

        class Scope(object):
            RT_SCOPE_UNIVERSE = 0x00  # Global route
            RT_SCOPE_SITE = 0x00  # Interior route in the local autonomous system
            RT_SCOPE_LINK = 0x00  # Route on this link
            RT_SCOPE_HOST = 0x00  # Route on the local host
            RT_SCOPE_NOWHERE = 0x00  # Destination does not exist

        class Attribute(object):
            IFLA_UNSPEC = 0x00
            IFLA_ADDRESS = 0x01
            IFLA_BROADCAST = 0x02
            IFLA_IFNAME = 0x03
            IFLA_MTU = 0x04
            IFLA_LINK = 0x05
            IFLA_QDISC = 0x06
            IFLA_STATS = 0x07
            IFLA_COST = 0x08
            IFLA_PRIORITY = 0x09
            IFLA_MASTER = 0x0A
            IFLA_WIRELESS = 0x0B
            IFLA_PROTINFO = 0x0C
            IFLA_TXQLEN = 0x0D
            IFLA_MAP = 0x0E
            IFLA_WEIGHT = 0x0F
            IFLA_OPERSTATE = 0x10
            IFLA_LINKMODE = 0x11
            IFLA_LINKINFO = 0x12
            IFLA_NET_NS_PID = 0x13
            IFLA_IFALIAS = 0x14
            IFLA_NUM_VF = 0x15
            IFLA_VFINFO_LIST = 0x16
            IFLA_STATS64 = 0x17
            IFLA_VF_PORTS = 0x18
            IFLA_PORT_SELF = 0x19

    @classmethod
    def getAddresses(cls):
        return cls.extract(Address.Command.RTM_GETADDR)



# encoding: utf-8
"""
nunrsvpbw.py

Created by Evelio Vila on 2016-12-01.
Copyright (c) 2014-2017 Exa Networks. All rights reserved.
"""

from struct import unpack

from exabgp.bgp.message.notification import Notify

from exabgp.bgp.message.update.attribute.bgpls.linkstate import LINKSTATE


#   This sub-TLV contains the amount of bandwidth reservable in this
#   direction on this link.  Note that for oversubscription purposes,
#   this can be greater than the bandwidth of the link.
#    [ One value per priority]
#   https://tools.ietf.org/html/rfc5305#section-3.6
#
#  Units are in Bytes not Bits.
#  ----------------------------


@LINKSTATE.register()
class UnRsvpBw(object):
    TLV = 1091

    def __init__(self, unrsvpbw):
        self.unrsvpbw = unrsvpbw

    def __repr__(self):
        return "Maximum link bandwidth: %s" % (self.unrsvpbw)

    @classmethod
    def unpack(cls, data, length):
        if length != 32:
            raise Notify(3, 5, "Wrong Unreservable Bw metric size")
        else:
            unrsvpbw = [p for p in unpack('!ffffffff', data)]
            return cls(unrsvpbw=unrsvpbw)

    def json(self, compact=None):
        return '"unreserved-bandwidth": %s' % str(self.unrsvpbw)



# encoding: utf-8
"""
eor.py

Created by Thomas Mangin on 2010-01-16.
Copyright (c) 2009-2017 Exa Networks. All rights reserved.
License: 3-clause BSD. (See the COPYRIGHT file)
"""

# from struct import unpack
import sys

from exabgp.util import character

from exabgp.protocol.family import AFI
from exabgp.protocol.family import SAFI
from exabgp.bgp.message.message import Message
from exabgp.bgp.message.update.attribute import Attributes
from exabgp.bgp.message.update.nlri import NLRI as _NLRI

if sys.version_info > (3,):
    long = int

# =================================================================== End-Of-RIB
# not technically a different message type but easier to treat as one


class EOR(Message):
    ID = Message.CODE.UPDATE
    TYPE = character(Message.CODE.UPDATE)

    class NLRI(_NLRI):
        PREFIX = b'\x00\x00\x00\x07\x90\x0F\x00\x03'
        MP_LENGTH = len(PREFIX) + 1 + 2  # len(AFI) and len(SAFI)
        EOR = True

        nexthop = None

        def __init__(self, afi, safi, action):
            _NLRI.__init__(self, afi, safi, action)
            self.action = action
            self.afi = afi
            self.safi = safi

        def pack_nlri(self, negotiated=None):
            if self.afi == AFI.ipv4 and self.safi == SAFI.unicast:
                return b'\x00\x00\x00\x00'
            return self.PREFIX + self.afi.pack() + self.safi.pack()

        def __repr__(self):
            return self.extensive()

        def extensive(self):
            return 'eor %ld/%ld (%s %s)' % (long(self.afi), long(self.safi), self.afi, self.safi)

        def json(self, announced=True, compact=None):
            return '"eor": { "afi" : "%s", "safi" : "%s" }' % (self.afi, self.safi)

        def __len__(self):
            if self.afi == AFI.ipv4 and self.safi == SAFI.unicast:
                # May not have been the size read on the wire if MP was used for IPv4 unicast
                return 4
            else:
                return self.MP_LENGTH

    def __init__(self, afi, safi, action=None):
        Message.__init__(self)
        self.nlris = [
            EOR.NLRI(afi, safi, action),
        ]
        self.attributes = Attributes()

    def message(self, negotiated=None):
        return self._message(self.nlris[0].pack())

    def __repr__(self):
        return 'EOR'

    @classmethod
    def unpack_message(cls, data, direction, negotiated):
        header_length = len(EOR.NLRI.PREFIX)
        return cls(AFI.unpack(data[header_length : header_length + 2]), SAFI.unpack(data[header_length + 2]))



# encoding: utf-8
"""
origin.py

Created by Thomas Mangin on 2009-11-05.
Copyright (c) 2009-2017 Exa Networks. All rights reserved.
License: 3-clause BSD. (See the COPYRIGHT file)
"""

from exabgp.util import character
from exabgp.util import ordinal
from exabgp.bgp.message.update.attribute.attribute import Attribute


# =================================================================== Origin (1)


@Attribute.register()
class Origin(Attribute):
    ID = Attribute.CODE.ORIGIN
    FLAG = Attribute.Flag.TRANSITIVE
    CACHING = True

    IGP = 0x00
    EGP = 0x01
    INCOMPLETE = 0x02

    __slots__ = ['origin', '_packed']

    def __init__(self, origin, packed=None):
        self.origin = origin
        self._packed = self._attribute(packed if packed else character(origin))

    def __eq__(self, other):
        return self.ID == other.ID and self.FLAG == other.FLAG and self.origin == other.origin

    def __ne__(self, other):
        return not self.__eq__(other)

    def pack(self, negotiated=None):
        return self._packed

    def __len__(self):
        return len(self._packed)

    def __repr__(self):
        if self.origin == 0x00:
            return 'igp'
        if self.origin == 0x01:
            return 'egp'
        if self.origin == 0x02:
            return 'incomplete'
        return 'invalid'

    @classmethod
    def unpack(cls, data, direction, negotiated):
        return cls(ordinal(data), data)

    @classmethod
    def setCache(cls):
        # there can only be three, build them now
        IGP = Origin(Origin.IGP)
        EGP = Origin(Origin.EGP)
        INC = Origin(Origin.INCOMPLETE)

        cls.cache[Attribute.CODE.ORIGIN][IGP.pack()] = IGP
        cls.cache[Attribute.CODE.ORIGIN][EGP.pack()] = EGP
        cls.cache[Attribute.CODE.ORIGIN][INC.pack()] = INC


Origin.setCache()



# encoding: utf-8
"""
prefixmetric.py

Created by Evelio Vila on 2016-12-01.
Copyright (c) 2014-2017 Exa Networks. All rights reserved.
"""

from struct import unpack

from exabgp.vendoring.bitstring import BitArray
from exabgp.bgp.message.notification import Notify

from exabgp.bgp.message.update.attribute.bgpls.linkstate import LINKSTATE

#
#      0                   1                   2                   3
#      0 1 2 3 4 5 6 7 8 9 0 1 2 3 4 5 6 7 8 9 0 1 2 3 4 5 6 7 8 9 0 1
#     +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
#     |              Type             |             Length            |
#     +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
#     |                            Metric                             |
#     +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
#     https://tools.ietf.org/html/rfc7752#section-3.3.3.4


@LINKSTATE.register()
class PrefixMetric(object):
    TLV = 1155

    def __init__(self, prefixmetric):
        self.prefixmetric = prefixmetric

    def __repr__(self):
        return "prefix_metric: %s" % (self.prefixmetric)

    @classmethod
    def unpack(cls, data, length):
        if length != 4:
            raise Notify(3, 5, "Incorrect Prefix Metric size")
        else:
            metric = unpack("!L", data)[0]
            return cls(prefixmetric=metric)

    def json(self, compact=None):
        return '"prefix-metric": %d' % int(self.prefixmetric)



# encoding: utf-8
"""
operational.py

Created by Thomas Mangin on 2013-09-01.
Copyright (c) 2013-2017 Exa Networks. All rights reserved.
License: 3-clause BSD. (See the COPYRIGHT file)
"""

from exabgp.bgp.message.open.capability.capability import Capability

# https://tools.ietf.org/html/draft-ietf-idr-operational-message-00
# ================================================================== Operational
#


@Capability.register()
class Operational(Capability, list):
    ID = Capability.CODE.OPERATIONAL

    def __str__(self):
        # XXX: FIXME: could be more verbose
        return 'Operational'

    def json(self):
        return '{ "name": "operational" }'

    def extract(self):
        return [b'']

    @staticmethod
    def unpack_capability(instance, data, capability=None):  # pylint: disable=W0613
        # XXX: FIXME: we should set that that instance was seen and raise if seen twice
        return instance



# encoding: utf-8
"""
community.py

Created by Thomas Mangin on 2009-11-05.
Copyright (c) 2009-2017 Exa Networks. All rights reserved.
License: 3-clause BSD. (See the COPYRIGHT file)
"""

from exabgp.util import ordinal
from exabgp.bgp.message.update.attribute import Attribute

from struct import pack

# ======================================================= ExtendedCommunity (16)
# XXX: Should subclasses register with transitivity ?


class ExtendedCommunityBase(Attribute):
    COMMUNITY_TYPE = 0x00  # MUST be redefined by subclasses
    COMMUNITY_SUBTYPE = 0x00  # MUST be redefined by subclasses
    NON_TRANSITIVE = 0x40

    # Need to be overwritten by sub-classes
    registered_extended = None

    @classmethod
    def register(cls, klass):
        cls.registered_extended[(klass.COMMUNITY_TYPE & 0x0F, klass.COMMUNITY_SUBTYPE)] = klass
        return klass

    # size of value for data (boolean: is extended)
    length_value = {False: 7, True: 6}
    name = {False: 'regular', True: 'extended'}

    __slots__ = ['community']

    def __init__(self, community):
        # Two top bits are iana and transitive bits
        self.community = community
        self.klass = None

    def __eq__(self, other):
        return self.ID == other.ID and self.FLAG == other.FLAG and self.community == other.community

    def __ne__(self, other):
        return not self.__eq__(other)

    def __lt__(self, other):
        return self.community < other.community

    def __le__(self, other):
        return self.community <= other.community

    def __gt__(self, other):
        return self.community > other.community

    def __ge__(self, other):
        return self.community >= other.community

    def iana(self):
        return not not (self.community[0] & 0x80)

    def transitive(self):
        # bit set means "not transitive"
        # RFC4360:
        #   T - Transitive bit
        #     Value 0: The community is transitive across ASes
        #     Value 1: The community is non-transitive across ASes
        return not (self.community[0] & 0x40)

    def pack(self, negotiated=None):
        return self.community

    def _subtype(self, transitive=True):
        # if not transitive -> set the 'transitive' bit, as per RFC4360
        return pack(
            '!BB',
            self.COMMUNITY_TYPE if transitive else self.COMMUNITY_TYPE | self.NON_TRANSITIVE,
            self.COMMUNITY_SUBTYPE,
        )

    def json(self):
        h = 0x00
        for byte in self.community:
            h <<= 8
            h += ordinal(byte)
        s = self.klass.__repr__(self) if self.klass else ''
        return '{ "value": %s, "string": "%s" }' % (h, s)

    def __repr__(self):
        if self.klass:
            return self.klass.__repr__(self)
        h = 0x00
        for byte in self.community:
            h <<= 8
            h += ordinal(byte)
        return "0x%016X" % h

    def __hash__(self):
        return hash(self.community)

    @classmethod
    def unpack(cls, data, direction=None, negotiated=None):
        # 30/02/12 Quagga communities for soo and rt are not transitive when 4360 says they must be, hence the & 0x0FFF
        community = (ordinal(data[0]) & 0x0F, ordinal(data[1]))
        if community in cls.registered_extended:
            klass = cls.registered_extended[community]
            instance = klass.unpack(data)
            instance.klass = klass
            return instance
        return cls(data)


class ExtendedCommunity(ExtendedCommunityBase):
    ID = Attribute.CODE.EXTENDED_COMMUNITY
    FLAG = Attribute.Flag.TRANSITIVE | Attribute.Flag.OPTIONAL

    registered_extended = {}

    def __len__(self):
        return 8


class ExtendedCommunityIPv6(ExtendedCommunityBase):
    ID = Attribute.CODE.IPV6_EXTENDED_COMMUNITY
    FLAG = Attribute.Flag.TRANSITIVE | Attribute.Flag.OPTIONAL

    registered_extended = {}

    def __len__(self):
        return 20



# encoding: utf-8
"""
unknown.py

Created by Thomas Mangin on 2013-07-20.
Copyright (c) 2009-2017 Exa Networks. All rights reserved.
License: 3-clause BSD. (See the COPYRIGHT file)
"""

from exabgp.util import character

from exabgp.bgp.message.message import Message

# ================================================================= UnknownMessage
#


class UnknownMessage(Message):
    # Make sure we have a value, which is not defined in any RFC !

    def __init__(self, code, data=b''):
        self.ID = code
        self.TYPE = character(code)
        self.data = data

    def message(self, negotiated=None):
        return self._message(self.data)

    def __str__(self):
        return "UNKNOWN"

    @classmethod
    def unpack_message(cls, data, direction=None, negotiated=None):  # pylint: disable=W0613
        raise RuntimeError('should not have been used')


UnknownMessage.klass_unknown = UnknownMessage



# encoding: utf-8
"""
iso

Created by Evelio Vila on 2016-11-26. eveliovila@gmail.com
Copyright (c) 2009-2017 Exa Networks. All rights reserved.
"""


# =========================================================================== ISO
#

from exabgp.vendoring.bitstring import BitArray


class ISO(object):
    def __init__(self, sysid, selector=None, area_id=None, afi=49):

        self.sysid = sysid
        self.area_id = area_id
        self.selector = selector
        self.afi = afi

    @classmethod
    def unpack_sysid(cls, data):
        b = BitArray(bytes=data)
        return b.hex

    def json(self, compact=None):
        return self.sysid



#!/usr/local/bin/python3.13

import os
import sys
import time


def _prefixed(level, message):
    now = time.strftime('%a, %d %b %Y %H:%M:%S\n', time.localtime())
    return '%s %-8s %-6d %s\n' % (now, level, os.getpid(), message)


# When the parent dies we are seeing continual newlines, so we only access so many before stopping
counter = 0


while True:
    try:
        line = sys.stdin.readline().strip()
        sys.stdout.flush()
        if line == "":
            counter += 1
            if counter > 100:
                break
            continue

        counter = 0

        sys.stderr.write(_prefixed(sys.argv[1] if len(sys.argv) >= 2 else 'EXABGP PROCESS', line))
    except KeyboardInterrupt:
        pass
    except IOError:
        # most likely a signal during readline
        pass



# encoding: utf-8
"""
sequence.py

Created by Thomas Mangin on 2015-03-31.
Copyright (c) 2009-2017 Exa Networks. All rights reserved.
License: 3-clause BSD. (See the COPYRIGHT file)
"""


class Sequence(int):
    _instance = dict()

    def __new__(cls):
        cls._instance['next'] = cls._instance.get('next', 0) + 1
        return cls._instance['next']



# encoding: utf-8
"""
clusterlist.py

Created by Thomas Mangin on 2012-07-07.
Copyright (c) 2009-2017 Exa Networks. All rights reserved.
License: 3-clause BSD. (See the COPYRIGHT file)
"""

from exabgp.util import concat_bytes_i

from exabgp.protocol.ip import IPv4

from exabgp.bgp.message.update.attribute.attribute import Attribute


# ===================================================================
#


class ClusterID(IPv4):
    def __init__(self, ip):
        IPv4.__init__(self, ip)


@Attribute.register()
class ClusterList(Attribute):
    ID = Attribute.CODE.CLUSTER_LIST
    FLAG = Attribute.Flag.OPTIONAL
    CACHING = True

    __slots__ = ['clusters', 'packed', '_len']

    def __init__(self, clusters, packed=None):
        self.clusters = clusters
        self._packed = self._attribute(packed if packed else concat_bytes_i(_.pack() for _ in clusters))
        self._len = len(clusters) * 4

    def __eq__(self, other):
        return self.ID == other.ID and self.FLAG == other.FLAG and self.clusters == other.clusters

    def __ne__(self, other):
        return not self.__eq__(other)

    def pack(self, negotiated=None):
        return self._packed

    def __len__(self):
        return self._len

    def __repr__(self):
        if self._len != 1:
            return '[ %s ]' % ' '.join([str(_) for _ in self.clusters])
        return '%s' % self.clusters[0]

    def json(self):
        return '[ %s ]' % ', '.join(['"%s"' % str(_) for _ in self.clusters])

    @classmethod
    def unpack(cls, data, direction, negotiated):
        clusters = []
        while data:
            clusters.append(IPv4.unpack(data[:4]))
            data = data[4:]
        return cls(clusters)



# encoding: utf-8
"""
logger.py

Created by Thomas Mangin on 2009-09-06.
Copyright (c) 2009-2017 Exa Networks. All rights reserved.
License: 3-clause BSD. (See the COPYRIGHT file)
"""

from __future__ import print_function

import os
import sys
import stat
import time
import syslog
import logging
import logging.handlers
import pdb
from collections import deque

from exabgp.util.od import od
from exabgp.configuration.environment import environment


def _can_write(location):
    try:
        s = os.stat(os.path.dirname(location))
    except OSError:
        return None
    mode = s[stat.ST_MODE]
    uid = os.geteuid()
    gid = os.getegid()

    return not not (
        ((s[stat.ST_UID] == uid) and (mode & stat.S_IWUSR))
        or ((s[stat.ST_GID] == gid) and (mode & stat.S_IWGRP))
        or (mode & stat.S_IWOTH)
    )


# This delays the evaluation of the od() function which is expensive
# hence why pylint too-few-public-method is disabled


class LazyFormat(object):
    def __init__(self, prefix, message, formater=od):
        self.prefix = prefix
        self.message = message
        self.formater = formater

    def split(self, char):
        return str(self).split(char)

    def __str__(self):
        formated = self.formater(self.message)
        return '%s (%4d) %s' % (self.prefix, len(self.message), formated)


class LazyAttribute(object):
    def __init__(self, flag, aid, length, data):
        self.flag = flag
        self.aid = aid
        self.length = length
        self.data = data

    def split(self, char):
        return str(self).split(char)

    def __str__(self):
        return 'attribute %-18s flag 0x%02x type 0x%02x len 0x%02x%s' % (
            str(self.aid),
            self.flag,
            int(self.aid),
            self.length,
            ' payload %s' % od(self.data) if self.data else '',
        )


class LazyNLRI(object):
    def __init__(self, afi, safi, addpath, data):
        self.afi = afi
        self.safi = safi
        self.addpath = addpath
        self.data = data

    def split(self, char):
        return str(self).split(char)

    def __str__(self):
        family = '%s %s' % (self.afi, self.safi)
        path = 'with path-information' if self.addpath else 'without path-information'
        payload = od(self.data) if self.data else 'none'
        return 'NLRI      %-18s %-28s payload %s' % (family, path, payload)


def istty(std):
    try:
        return std.isatty()
    except KeyboardInterrupt:
        raise
    except Exception:
        return False


class Logger(object):
    RECORD = {
        'START': '\033[01;32m',  # Green
        'DEBUG': '',
        'INFO': '\033[01;32m',  # Green
        'NOTICE': '\033[01;34m',  # Blue
        'WARNING': '\033[01;33m',  # Yellow
        'ERR': '\033[01;31m',  # Red
        'CRIT': '\033[00;31m',  # Strong Red
    }

    MESSAGE = {
        'START': '\033[1m',  # Green
        'DEBUG': '',
        'INFO': '\033[1m',  # Green
        'NOTICE': '\033[1m',  # Blue
        'WARNING': '\033[1m',  # Yellow
        'ERR': '\033[1m',  # Red
        'CRIT': '\033[1m',  # Strong Red
    }

    END = '\033[0m'

    TTY = {
        'stderr': lambda: istty(sys.stderr),
        'stdout': lambda: istty(sys.stdout),
        'out': lambda: istty(sys.stdout),
    }

    _instance = dict()
    _syslog = None
    _where = ''

    _history = deque()
    _max_history = 20

    _default = None

    _config = ''
    _pid = os.getpid()
    _cwd = os.getcwd()

    def __new__(cls):
        if cls._instance.get('class', None) is None:
            return super(Logger, cls).__new__(cls)
        else:
            return cls._instance['class']

    # we use os.pid everytime as we may fork and the class is instance before it

    def pdb(self, level):
        if self._option['pdb'] and level == 'CRIT':
            # not sure why, pylint reports an import error here
            pdb.set_trace()

    def config(self, config=None):
        if config is not None:
            self._config = config
        return self._config

    def history(self):
        return "\n".join(self._format(*_) for _ in self._history)

    def _record(self, timestamp, message, source, level):
        if len(self._history) > self._max_history:
            self._history.popleft()
        self._history.append((message, source, level, timestamp))

    def __init__(self):
        if self._instance.get('class', None) is not None:
            return

        self._instance['class'] = self

        command = environment.settings()
        self.short = command.log.short
        self.level = command.log.level

        self._option = {
            'pdb': command.debug.pdb,
            'reactor': command.log.enable and (command.log.all or command.log.reactor),
            'daemon': command.log.enable and (command.log.all or command.log.daemon),
            'processes': command.log.enable and (command.log.all or command.log.processes),
            'configuration': command.log.enable and (command.log.all or command.log.configuration),
            'network': command.log.enable and (command.log.all or command.log.network),
            'wire': command.log.enable and (command.log.all or command.log.packets),
            'message': command.log.enable and (command.log.all or command.log.message),
            'rib': command.log.enable and (command.log.all or command.log.rib),
            'timer': command.log.enable and (command.log.all or command.log.timers),
            'routes': command.log.enable and (command.log.all or command.log.routes),
            'parser': command.log.enable and (command.log.all or command.log.parser),
        }

        if not command.log.enable:
            self.destination = ''
            return

        self.destination = command.log.destination

        self.restart(True)

    def _local_syslog(self):
        if sys.platform == "darwin":
            address = '/var/run/syslog'
        else:
            address = '/dev/log'
        if not os.path.exists(address):
            address = ('localhost', 514)
        handler = logging.handlers.SysLogHandler(address)

        self._syslog.addHandler(handler)
        return True

    def _remote_syslog(self, destination):
        # If the address is invalid, each syslog call will print an error.
        # See how it can be avoided, as the socket error is encapsulated and not returned
        address = (destination, 514)
        handler = logging.handlers.SysLogHandler(address)

        self._syslog.addHandler(handler)
        return True

    def _standard(self, facility):
        handler = logging.StreamHandler(getattr(sys, facility))

        self._syslog.addHandler(handler)
        return True

    def _file(self, destination):
        # folder
        logfile = os.path.realpath(os.path.normpath(os.path.join(self._cwd, destination)))
        can = _can_write(logfile)
        if can is True:
            handler = logging.handlers.RotatingFileHandler(logfile, maxBytes=5 * 1024 * 1024, backupCount=5)
        elif can is None:
            self.critical('ExaBGP can not access (perhaps as it does not exist) the log folder provided', 'logger')
            return False
        else:
            self.critical('ExaBGP does not have the right to write in the requested log directory', 'logger')
            return False

        self._syslog.addHandler(handler)
        return True

    def restart(self, first=False):
        try:
            if first:
                self._where = 'stdout'
                self._default = logging.StreamHandler(sys.stdout)
                self._syslog = logging.getLogger()
                self._syslog.setLevel(logging.DEBUG)
                self._syslog.addHandler(self._default)
                return True
        except IOError:
            # no way to report anything via stdout, silently failing
            return False

        if not self._syslog:
            # no way to report anything via stdout, silently failing
            return False

        for handler in self._syslog.handlers:
            self._syslog.removeHandler(handler)
        self._syslog.addHandler(self._default)

        try:
            if self.destination == 'stderr':
                self._where = 'stderr'
                return True
            elif self.destination == 'stdout':
                self._where = 'out'
                result = self._standard(self.destination)
            elif self.destination in ('', 'syslog'):
                self._where = 'syslog'
                result = self._local_syslog()
            elif self.destination.startswith('host:'):
                self._where = 'syslog'
                result = self._remote_syslog(self.destination[5:].strip())
            else:
                self._where = 'file'
                result = self._file(self.destination)

            if result:
                self._syslog.removeHandler(self._default)
            return result
        except IOError:
            self.critical('Can not set logging (are stdout/stderr closed?)', 'logger')
            return False

    def _format(self, message, source, level, timestamp=None):
        if timestamp is None:
            timestamp = time.localtime()
            self._record(timestamp, message, source, level)

        if self.short:
            return message

        if self._where in ['stdout', 'stderr', 'out']:
            now = time.strftime('%H:%M:%S', timestamp)
            if not self.TTY[self._where]():
                return "%s | %-6d | %-15s | %s" % (now, self._pid, source, message)
            return "%s | %-6d | %s%-13s%s | %s%-8s%s" % (
                now,
                self._pid,
                self.RECORD.get(level, ''),
                source,
                self.END,
                self.MESSAGE.get(level, ''),
                message,
                self.END,
            )
        elif self._where in [
            'syslog',
        ]:
            return "%s[%d]: %-13s %s" % (environment.application, self._pid, source, message)
        elif self._where in [
            'file',
        ]:
            now = time.strftime('%a, %d %b %Y %H:%M:%S', timestamp)
            return "%s %-6d %-13s %s" % (now, self._pid, source, message)
        else:
            # failsafe
            return "%s | %-8s | %-6d | %-13s | %s" % (now, level, self._pid, source, message)

    def _report(self, message, source, level):
        if source.startswith('incoming-'):
            src = 'wire'
        elif source.startswith('outgoing-'):
            src = 'wire'
        elif source.startswith('ka-'):
            src = 'timer'
        elif source.startswith('peer-'):
            src = 'network'
        else:
            src = source

        log = self._option.get(src, True) and getattr(syslog, 'LOG_%s' % level) <= self.level

        if not log:
            return

        for line in message.split('\n'):
            if self._syslog:
                self._syslog.debug(self._format(line, source, level))
            else:
                print(self._format(line, source, level))
                sys.stdout.flush()

    def debug(self, message, source='', level='DEBUG'):
        self._report(message, source, level)

    def info(self, message, source='', level='INFO'):
        self._report(message, source, level)

    def notice(self, message, source='', level='NOTICE'):
        self._report(message, source, level)

    def warning(self, message, source='', level='WARNING'):
        self._report(message, source, level)

    def error(self, message, source='', level='ERR'):
        self._report(message, source, level)

    def critical(self, message, source='', level='CRIT'):
        self._report(message, source, level)


class FakeLogger(object):
    def __getattr__(self, name):
        def printf(data, _=None):
            sys.stdout.write('Fake logger [%s]\n' % str(data))

        return printf



# encoding: utf-8
"""
neighbor.py

Created by Thomas Mangin on 2009-11-05.
Copyright (c) 2009-2017 Exa Networks. All rights reserved.
License: 3-clause BSD. (See the COPYRIGHT file)
"""

from copy import deepcopy

from collections import deque
from collections import Counter

from exabgp.protocol.family import AFI

from exabgp.bgp.message import Message
from exabgp.bgp.message.update.attribute import NextHop
from exabgp.bgp.message.update.attribute import Attribute

from exabgp.rib import RIB


# The definition of a neighbor (from reading the configuration)
class Neighbor(object):
    _GLOBAL = {'uid': 1}

    def __init__(self):
        # self.logger should not be used here as long as we do use deepcopy as it contains a Lock
        self.description = None
        self.router_id = None
        self.host_name = None
        self.domain_name = None
        self.local_address = None
        self.range_size = 1
        # local_address uses auto discovery
        self.auto_discovery = False
        self.peer_address = None
        self.peer_as = None
        self.local_as = None
        self.hold_time = None
        self.rate_limit = None
        self.asn4 = None
        self.nexthop = None
        self.add_path = None
        self.md5_password = None
        self.md5_base64 = False
        self.md5_ip = None
        self.ttl_in = None
        self.ttl_out = None
        self.group_updates = None
        self.flush = None
        self.adj_rib_in = None
        self.adj_rib_out = None

        self.manual_eor = False

        self.api = None  # XXX: not scriptable - is replaced outside the class
        # passive indicate that we do not establish outgoing connections
        self.passive = False
        # the port to listen on ( zero mean that we do not listen )
        self.listen = 0
        # the port to connect to
        self.connect = 0

        # was this Neighbor generated from a range
        self.generated = False

        # capability
        self.route_refresh = False
        self.graceful_restart = False
        self.multisession = None
        self.nexthop = None
        self.add_path = None
        self.aigp = None

        self._families = []
        self._nexthop = []
        self._addpath = []
        self.rib = None

        # The routes we have parsed from the configuration
        self.changes = []

        self.operational = None
        self.eor = deque()
        self.asm = dict()

        self.messages = deque()
        self.refresh = deque()

        self.counter = Counter()
        # It is possible to :
        # - have multiple exabgp toward one peer on the same host ( use of pid )
        # - have more than once connection toward a peer
        # - each connection has it own neihgbor (hence why identificator is not in Protocol)
        self.uid = '%d' % self._GLOBAL['uid']
        self._GLOBAL['uid'] += 1

    def id(self):
        return 'neighbor-%s' % self.uid

    # This set must be unique between peer, not full draft-ietf-idr-bgp-multisession-07
    def index(self):
        if self.listen != 0:
            return 'peer-ip %s listen %d' % (self.peer_address, self.listen)
        return self.name()

    def make_rib(self):
        self.rib = RIB(self.name(), self.adj_rib_in, self.adj_rib_out, self._families)

    # will resend all the routes once we reconnect
    def reset_rib(self):
        self.rib.reset()
        self.messages = deque()
        self.refresh = deque()

    # back to square one, all the routes are removed
    def clear_rib(self):
        self.rib.clear()
        self.messages = deque()
        self.refresh = deque()

    def name(self):
        if self.multisession:
            session = '/'.join("%s-%s" % (afi.name(), safi.name()) for (afi, safi) in self.families())
        else:
            session = 'in-open'
        return "neighbor %s local-ip %s local-as %s peer-as %s router-id %s family-allowed %s" % (
            self.peer_address,
            self.local_address if self.peer_address is not None else 'auto',
            self.local_as if self.local_as is not None else 'auto',
            self.peer_as if self.peer_as is not None else 'auto',
            self.router_id,
            session,
        )

    def families(self):
        # this list() is important .. as we use the function to modify self._families
        return list(self._families)

    def nexthops(self):
        # this list() is important .. as we use the function to modify self._nexthop
        return list(self._nexthop)

    def addpaths(self):
        # this list() is important .. as we use the function to modify self._add_path
        return list(self._addpath)

    def add_family(self, family):
        # the families MUST be sorted for neighbor indexing name to be predictable for API users
        # this list() is important .. as we use the function to modify self._families
        if family not in self.families():
            afi, safi = family
            d = dict()
            d[afi] = [
                safi,
            ]
            for afi, safi in self._families:
                d.setdefault(afi, []).append(safi)
            self._families = [(afi, safi) for afi in sorted(d) for safi in sorted(d[afi])]

    def add_nexthop(self, afi, safi, nhafi):
        if (afi, safi, nhafi) not in self._nexthop:
            self._nexthop.append((afi, safi, nhafi))

    def add_addpath(self, family):
        # the families MUST be sorted for neighbor indexing name to be predictable for API users
        # this list() is important .. as we use the function to modify self._add_path
        if family not in self.addpaths():
            afi, safi = family
            d = dict()
            d[afi] = [
                safi,
            ]
            for afi, safi in self._addpath:
                d.setdefault(afi, []).append(safi)
            self._addpath = [(afi, safi) for afi in sorted(d) for safi in sorted(d[afi])]

    def remove_family(self, family):
        if family in self.families():
            self._families.remove(family)

    def remove_nexthop(self, afi, safi, nhafi):
        if (afi, safi, nhafi) in self.nexthops():
            self._nexthop.remove((afi, safi, nhafi))

    def remove_addpath(self, family):
        if family in self.addpaths():
            self._addpath.remove(family)

    def process_previous_changes(self, previous_changes):
        self.rib.outgoing.replace(previous_changes, self.changes)

    def missing(self):
        if self.local_address is None and not self.auto_discovery:
            return 'local-address'
        if self.listen > 0 and self.auto_discovery:
            return 'local-address'
        if self.peer_address is None:
            return 'peer-address'
        if self.auto_discovery and not self.router_id:
            return 'router-id'
        if self.peer_address.afi == AFI.ipv6 and not self.router_id:
            return 'router-id'
        return ''

    # This function only compares the neighbor BUT NOT ITS ROUTES
    def __eq__(self, other):
        # Comparing local_address is skipped in the case where either
        # peer is configured to auto discover its local address. In
        # this case it can happen that one local_address is None and
        # the other one will be set to the auto disocvered IP address.
        auto_discovery = self.auto_discovery or other.auto_discovery
        return (
            self.router_id == other.router_id
            and (auto_discovery or self.local_address == other.local_address)
            and self.auto_discovery == other.auto_discovery
            and self.local_as == other.local_as
            and self.peer_address == other.peer_address
            and self.peer_as == other.peer_as
            and self.passive == other.passive
            and self.listen == other.listen
            and self.connect == other.connect
            and self.hold_time == other.hold_time
            and self.rate_limit == other.rate_limit
            and self.host_name == other.host_name
            and self.domain_name == other.domain_name
            and self.md5_password == other.md5_password
            and self.md5_ip == other.md5_ip
            and self.ttl_in == other.ttl_in
            and self.ttl_out == other.ttl_out
            and self.route_refresh == other.route_refresh
            and self.graceful_restart == other.graceful_restart
            and self.multisession == other.multisession
            and self.nexthop == other.nexthop
            and self.add_path == other.add_path
            and self.operational == other.operational
            and self.group_updates == other.group_updates
            and self.flush == other.flush
            and self.adj_rib_in == other.adj_rib_in
            and self.adj_rib_out == other.adj_rib_out
            and self.families() == other.families()
        )

    def __ne__(self, other):
        return not self.__eq__(other)

    def string(self, with_changes=True):
        changes = ''
        if with_changes:
            changes += '\nstatic { '
            for changes in self.rib.incoming.queued_changes():
                changes += '\n\t\t%s' % changes.extensive()
            changes += '\n}'

        families = ''
        for afi, safi in self.families():
            families += '\n\t\t%s %s;' % (afi.name(), safi.name())

        nexthops = ''
        for afi, safi, nexthop in self.nexthops():
            nexthops += '\n\t\t%s %s %s;' % (afi.name(), safi.name(), nexthop.name())

        addpaths = ''
        for afi, safi in self.addpaths():
            addpaths += '\n\t\t%s %s;' % (afi.name(), safi.name())

        codes = Message.CODE

        _extension_global = {
            'neighbor-changes': 'neighbor-changes',
            'negotiated': 'negotiated',
            'fsm': 'fsm',
            'signal': 'signal',
        }

        _extension_receive = {
            'receive-packets': 'packets',
            'receive-parsed': 'parsed',
            'receive-consolidate': 'consolidate',
            'receive-%s' % codes.NOTIFICATION.SHORT: 'notification',
            'receive-%s' % codes.OPEN.SHORT: 'open',
            'receive-%s' % codes.KEEPALIVE.SHORT: 'keepalive',
            'receive-%s' % codes.UPDATE.SHORT: 'update',
            'receive-%s' % codes.ROUTE_REFRESH.SHORT: 'refresh',
            'receive-%s' % codes.OPERATIONAL.SHORT: 'operational',
        }

        _extension_send = {
            'send-packets': 'packets',
            'send-parsed': 'parsed',
            'send-consolidate': 'consolidate',
            'send-%s' % codes.NOTIFICATION.SHORT: 'notification',
            'send-%s' % codes.OPEN.SHORT: 'open',
            'send-%s' % codes.KEEPALIVE.SHORT: 'keepalive',
            'send-%s' % codes.UPDATE.SHORT: 'update',
            'send-%s' % codes.ROUTE_REFRESH.SHORT: 'refresh',
            'send-%s' % codes.OPERATIONAL.SHORT: 'operational',
        }

        apis = ''

        for process in self.api.get('processes', []):
            _global = []
            _receive = []
            _send = []

            for api, name in _extension_global.items():
                _global.extend(['\t\t%s;\n' % name,] if process in self.api[api] else [])

            for api, name in _extension_receive.items():
                _receive.extend(['\t\t\t%s;\n' % name,] if process in self.api[api] else [])

            for api, name in _extension_send.items():
                _send.extend(['\t\t\t%s;\n' % name,] if process in self.api[api] else [])

            _api = '\tapi {\n'
            _api += '\t\tprocesses [ %s ];\n' % process
            _api += ''.join(_global)
            if _receive:
                _api += '\t\treceive {\n'
                _api += ''.join(_receive)
                _api += '\t\t}\n'
            if _send:
                _api += '\t\tsend {\n'
                _api += ''.join(_send)
                _api += '\t\t}\n'
            _api += '\t}\n'

            apis += _api

        returned = (
            'neighbor %s {\n'
            '\tdescription "%s";\n'
            '\trouter-id %s;\n'
            '\thost-name %s;\n'
            '\tdomain-name %s;\n'
            '\tlocal-address %s;\n'
            '\tlocal-as %s;\n'
            '\tpeer-as %s;\n'
            '\thold-time %s;\n'
            '\trate-limit %s;\n'
            '\tmanual-eor %s;\n'
            '%s%s%s%s%s%s%s%s%s%s%s\n'
            '\tcapability {\n'
            '%s%s%s%s%s%s%s%s%s\t}\n'
            '\tfamily {%s\n'
            '\t}\n'
            '\tnexthop {%s\n'
            '\t}\n'
            '\tadd-path {%s\n'
            '\t}\n'
            '%s'
            '%s'
            '}'
            % (
                self.peer_address,
                self.description,
                self.router_id,
                self.host_name,
                self.domain_name,
                self.local_address if not self.auto_discovery else 'auto',
                self.local_as,
                self.peer_as,
                self.hold_time,
                'disable' if self.rate_limit == 0 else self.rate_limit,
                'true' if self.manual_eor else 'false',
                '\n\tpassive %s;\n' % ('true' if self.passive else 'false'),
                '\n\tlisten %d;\n' % self.listen if self.listen else '',
                '\n\tconnect %d;\n' % self.connect if self.connect else '',
                '\tgroup-updates %s;\n' % ('true' if self.group_updates else 'false'),
                '\tauto-flush %s;\n' % ('true' if self.flush else 'false'),
                '\tadj-rib-in %s;\n' % ('true' if self.adj_rib_in else 'false'),
                '\tadj-rib-out %s;\n' % ('true' if self.adj_rib_out else 'false'),
                '\tmd5-password "%s";\n' % self.md5_password if self.md5_password else '',
                '\tmd5-base64 %s;\n'
                % ('true' if self.md5_base64 is True else 'false' if self.md5_base64 is False else 'auto'),
                '\tmd5-ip "%s";\n' % self.md5_ip if not self.auto_discovery else '',
                '\toutgoing-ttl %s;\n' % self.ttl_out if self.ttl_out else '',
                '\tincoming-ttl %s;\n' % self.ttl_in if self.ttl_in else '',
                '\t\tasn4 %s;\n' % ('enable' if self.asn4 else 'disable'),
                '\t\troute-refresh %s;\n' % ('enable' if self.route_refresh else 'disable'),
                '\t\tgraceful-restart %s;\n' % (self.graceful_restart if self.graceful_restart else 'disable'),
                '\t\tnexthop %s;\n' % ('enable' if self.nexthop else 'disable'),
                '\t\tadd-path %s;\n' % (AddPath.string[self.add_path] if self.add_path else 'disable'),
                '\t\tmulti-session %s;\n' % ('enable' if self.multisession else 'disable'),
                '\t\toperational %s;\n' % ('enable' if self.operational else 'disable'),
                '\t\taigp %s;\n' % ('enable' if self.aigp else 'disable'),
                families,
                nexthops,
                addpaths,
                apis,
                changes,
            )
        )

        # '\t\treceive {\n%s\t\t}\n' % receive if receive else '',
        # '\t\tsend {\n%s\t\t}\n' % send if send else '',
        return returned.replace('\t', '  ')

    def ip_self(self, afi):
        if afi == self.local_address.afi:
            return self.local_address

        # attempting to not barf for next-hop self when the peer is IPv6
        if afi == AFI.ipv4:
            return self.router_id

        raise TypeError(
            'use of "next-hop self": the route (%s) does not have the same family as the BGP tcp session (%s)'
            % (afi, self.local_address.afi)
        )

    def remove_self(self, changes):
        change = deepcopy(changes)
        if not change.nlri.nexthop.SELF:
            return change
        neighbor_self = self.ip_self(change.nlri.afi)
        change.nlri.nexthop = neighbor_self
        if Attribute.CODE.NEXT_HOP in change.attributes:
            change.attributes[Attribute.CODE.NEXT_HOP] = NextHop(str(neighbor_self),neighbor_self.pack())
        return change

    def __str__(self):
        return self.string(False)



# encoding: utf-8
"""
environment.py

Created by Thomas Mangin on 2011-11-29.
Copyright (c) 2011-2017 Exa Networks. All rights reserved.
License: 3-clause BSD. (See the COPYRIGHT file)
"""

# XXX: raised exception not caught
# XXX: reloading mid-program not possible
# XXX: validation for path, file, etc not correctly test (ie surely buggy)


import os
import sys
import pwd
import syslog

from exabgp.protocol.ip import IP
from exabgp.util.ip import isip


# ===================================================================== NoneDict
#


class NoneDict(dict):
    def __getitem__(self, name):
        return None


nonedict = NoneDict()


# ================================================================== environment
# XXX: FIXME: Upper case for class !


class environment(object):
    # class returned on issues
    class Error(Exception):
        pass

    application = 'unset'

    # the configuration to be set by the program
    configuration = {}

    # the final parsed settings
    _settings = None

    location = (
        os.path.normpath(sys.argv[0])
        if sys.argv[0].startswith('/')
        else os.path.normpath(os.path.join(os.getcwd(), sys.argv[0]))
    )
    log_levels = ['EMERG', 'ALERT', 'CRIT', 'CRITICAL', 'ERR', 'ERROR', 'WARNING', 'NOTICE', 'INFO', 'DEBUG']

    @staticmethod
    def setup(conf):
        if environment._settings:
            # nosetest is performing the setup multiple times, so we can not raise anymore
            # raise RuntimeError('You already initialised the environment')
            return environment._settings
        environment._settings = _env(conf)
        return environment._settings

    @staticmethod
    def settings():
        if not environment._settings:
            raise RuntimeError('You can not have an import using settings() before main() initialised environment')
        return environment._settings

    @staticmethod
    def root(path):
        roots = environment.location.split(os.sep)
        location = []
        for index in range(len(roots) - 1, -1, -1):
            if roots[index] == 'lib':
                if index:
                    location = roots[:index]
                break
        root = os.path.join(*location)
        paths = [
            os.path.normpath(os.path.join(os.path.join(os.sep, root, path))),
            os.path.normpath(os.path.expanduser(environment.unquote(path))),
            os.path.normpath(os.path.join('/', path)),
        ]
        return paths

    @staticmethod
    def integer(_):
        return int(_)

    @staticmethod
    def real(_):
        return float(_)

    @staticmethod
    def lowunquote(_):
        return _.strip().strip('\'"').lower()

    @staticmethod
    def unquote(_):
        return _.strip().strip('\'"')

    @staticmethod
    def quote(_):
        return "'%s'" % str(_)

    @staticmethod
    def quote_list(_):
        return "'%s'" % ' '.join([str(x) for x in _])

    @staticmethod
    def nop(_):
        return _

    @staticmethod
    def boolean(_):
        return _.lower() in ('1', 'yes', 'on', 'enable', 'true')

    @staticmethod
    def api(_):
        encoder = _.lower()
        if encoder not in ('text', 'json'):
            raise TypeError('invalid encoder')
        return encoder

    @staticmethod
    def methods(_):
        return _.upper().split()

    @staticmethod
    def list(_):
        return "'%s'" % ' '.join(_)

    @staticmethod
    def lower(_):
        return str(_).lower()

    @staticmethod
    def ip(_):
        if isip(_):
            return _
        raise TypeError('ip %s is invalid' % _)

    @staticmethod
    def ip_list(_):
        ips = []
        for ip in _.split(' '):
            if not ip:
                continue
            elif isip(ip):
                ips.append(IP.create(ip))
            else:
                raise TypeError('ip %s is invalid' % ip)
        return ips

    @staticmethod
    def user(_):
        # XXX: incomplete
        try:
            pwd.getpwnam(_)
            # uid = answer[2]
        except KeyError:
            raise TypeError('user %s is not found on this system' % _)
        return _

    @staticmethod
    def folder(path):
        paths = environment.root(path)
        options = [p for p in paths if os.path.exists(path)]
        if not options:
            raise TypeError('%s does not exists' % path)
        first = options[0]
        if not first:
            raise TypeError('%s does not exists' % first)
        return first

    @staticmethod
    def path(path):
        split = sys.argv[0].split('lib/exabgp')
        if len(split) > 1:
            prefix = os.sep.join(split[:1])
            if prefix and path.startswith(prefix):
                path = path[len(prefix) :]
        home = os.path.expanduser('~')
        if path.startswith(home):
            return "'~%s'" % path[len(home) :]
        return "'%s'" % path

    @staticmethod
    def conf(path):
        first = environment.folder(path)
        if not os.path.isfile(first):
            raise TypeError('%s is not a file' % path)
        return first

    @staticmethod
    def exe(path):
        first = environment.conf(path)
        if not os.access(first, os.X_OK):
            raise TypeError('%s is not an executable' % first)
        return first

    @staticmethod
    def syslog(path):
        path = environment.unquote(path)
        if path in ('stdout', 'stderr'):
            return path
        if path.startswith('host:'):
            return path
        return path

    @staticmethod
    def redirector(name):
        if name == 'url' or name.startswith('icap://'):
            return name
        raise TypeError('invalid redirector protocol %s, options are url or header' % name)

    @staticmethod
    def syslog_value(log):
        if log not in environment.log_levels:
            if log == 'CRITICAL':
                log = 'CRIT'
            if log == 'ERROR':
                log = 'ERR'
            raise TypeError('invalid log level %s' % log)
        return getattr(syslog, 'LOG_%s' % log)

    @staticmethod
    def syslog_name(log):
        for name in environment.log_levels:
            if name == 'CRITICAL':
                name = 'CRIT'
            if name == 'ERROR':
                name = 'ERR'
            if getattr(syslog, 'LOG_%s' % name) == log:
                return name
        raise TypeError('invalid log level %s' % log)

    @staticmethod
    def umask_read(_):
        return int(_, 8)

    @staticmethod
    def umask_write(_):
        return "'%s'" % (oct(_))

    @staticmethod
    def default():
        for section in sorted(environment.configuration):
            if section in ('internal', 'debug'):
                continue
            for option in sorted(environment.configuration[section]):
                values = environment.configuration[section][option]
                default = (
                    "'%s'" % values['value']
                    if values['write'] in (environment.list, environment.path, environment.quote, environment.syslog)
                    else values['value']
                )
                yield '%s.%s.%s %s: %s. default (%s)' % (
                    environment.application,
                    section,
                    option,
                    ' ' * (20 - len(section) - len(option)),
                    values['help'],
                    default,
                )

    @staticmethod
    def iter_ini(diff=False):
        for section in sorted(environment._settings):
            if section in ('internal', 'debug'):
                continue
            header = '\n[%s.%s]' % (environment.application, section)
            for k in sorted(environment._settings[section]):
                v = environment._settings[section][k]
                if (
                    diff
                    and environment.configuration[section][k]['read'](environment.configuration[section][k]['value'])
                    == v
                ):
                    continue
                if header:
                    yield header
                    header = ''
                yield '%s = %s' % (k, environment.configuration[section][k]['write'](v))

    @staticmethod
    def iter_env(diff=False):
        for section, values in environment._settings.items():
            if section in ('internal', 'debug'):
                continue
            for k, v in values.items():
                if (
                    diff
                    and environment.configuration[section][k]['read'](environment.configuration[section][k]['value'])
                    == v
                ):
                    continue
                if environment.configuration[section][k]['write'] == environment.quote:
                    yield "%s.%s.%s='%s'" % (environment.application, section, k, v)
                    continue
                yield "%s.%s.%s=%s" % (
                    environment.application,
                    section,
                    k,
                    environment.configuration[section][k]['write'](v),
                )


# ========================================================================= _env
#

from exabgp.vendoring.six.moves import configparser as ConfigParser
from exabgp.util.hashtable import HashTable


def _env(conf):
    here = os.path.join(os.sep, os.path.join(*environment.location.split(os.sep)))

    location, directory = os.path.split(here)
    while directory:
        if directory == 'lib':
            location = os.path.join(location, 'lib')
            break
        location, directory = os.path.split(location)
    # we did not break - ie, we did not find the location in the normal path.
    else:
        # let's try to see if we are running from the QA folder (for unittesting)
        location, directory = os.path.split(here)
        while directory:
            if directory == 'dev':
                location = os.path.join(location, 'lib')
                break
            location, directory = os.path.split(location)
        else:
            # oh ! bad, let set the path to something ...
            location = '/lib'

    _conf_paths = []
    if conf:
        _conf_paths.append(os.path.abspath(os.path.normpath(conf)))
    if location:
        _conf_paths.append(
            os.path.normpath(os.path.join(location, 'etc', environment.application, '%s.env' % environment.application))
        )
    _conf_paths.append(
        os.path.normpath(os.path.join('/', 'etc', environment.application, '%s.env' % environment.application))
    )

    env = HashTable()
    ini = ConfigParser.ConfigParser()

    ini_files = [path for path in _conf_paths if os.path.exists(path)]
    if ini_files:
        ini.read(ini_files[0])

    for section in environment.configuration:
        default = environment.configuration[section]

        for option in default:
            convert = default[option]['read']
            try:
                proxy_section = '%s.%s' % (environment.application, section)
                env_name = '%s.%s' % (proxy_section, option)
                rep_name = env_name.replace('.', '_')

                if env_name in os.environ:
                    conf = os.environ.get(env_name)
                elif rep_name in os.environ:
                    conf = os.environ.get(rep_name)
                else:
                    if sys.version_info[0] < 3:
                        conf = environment.unquote(ini.get(proxy_section, option, nonedict))
                    else:
                        conf = environment.unquote(ini.get(proxy_section, option, vars=nonedict))
                    # name without an = or : in the configuration and no value
                    if conf is None:
                        conf = default[option]['value']
            except (ConfigParser.NoSectionError, ConfigParser.NoOptionError):
                conf = default[option]['value']
            try:
                env.setdefault(section, HashTable())[option] = convert(conf)
            except TypeError:
                raise environment.Error('invalid value for %s.%s : %s' % (section, option, conf))

    return env



# encoding: utf-8
"""
nodename.py

Created by Evelio Vila on 2016-12-01.
Copyright (c) 2014-2017 Exa Networks. All rights reserved.
"""

from struct import pack
from struct import unpack

from exabgp.bgp.message.notification import Notify

from exabgp.bgp.message.update.attribute.bgpls.linkstate import LINKSTATE

#
#      0 1 2 3 4 5 6 7 8 9 0 1 2 3 4 5 6 7 8 9 0 1 2 3 4 5 6 7 8 9 0 1
#     +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
#     |              Type             |             Length            |
#     +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
#     //              Opaque Prefix Attributes  (variable)           //
#     +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
#     https://tools.ietf.org/html/rfc7752#section-3.3.3.6
#


@LINKSTATE.register()
class PrefixOpaque(object):
    TLV = 1157

    def __init__(self, opaque):
        self.opaque = opaque

    def __repr__(self):
        return "Prefix Opaque attribute: %s" % (self.opaque)

    @classmethod
    def unpack(cls, data, length):
        opaque = unpack("!%ds" % length, data[:length])[0]
        return cls(opaque=opaque)



# encoding: utf-8
"""
neighbor.py

Created by Thomas Mangin on 2015-03-31.
Copyright (c) 2009-2017 Exa Networks. All rights reserved.
License: 3-clause BSD. (See the COPYRIGHT file)
"""

import socket
from struct import calcsize
from collections import namedtuple

from exabgp.netlink.message import Message


# 0                   1                   2                   3
# 0 1 2 3 4 5 6 7 8 9 0 1 2 3 4 5 6 7 8 9 0 1 2 3 4 5 6 7 8 9 0 1
# +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
# |   Family    |    Reserved1  |           Reserved2           |
# +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
# |                     Interface Index                         |
# +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
# |           State             |     Flags     |     Type      |
# +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+


class Neighbor(Message):
    class Header(object):
        # linux/if_addr.h
        PACK = 'BxxxiHBB'
        LEN = calcsize(PACK)

    format = namedtuple('Neighbor', 'family index state flags type attributes')

    class Command(object):
        RTM_NEWNEIGH = 0x1C
        RTM_DELNEIGH = 0x1D
        RTM_GETNEIGH = 0x1E

    class Type(object):
        class Family(object):
            AF_INET = socket.AF_INET
            AF_INET6 = socket.AF_INET6

        class State(object):
            NUD_INCOMPLETE = 0x01  # Still attempting to resolve
            NUD_REACHABLE = 0x02  # A confirmed working cache entry
            NUD_STALE = 0x04  # an expired cache entry
            NUD_DELAY = 0x08  # Neighbor no longer reachable.  Traffic sent, waiting for confirmatio.
            NUD_PROBE = 0x10  # A cache entry that is currently being re-solicited
            NUD_FAILED = 0x20  # An invalid cache entry
            # Dummy states
            NUD_NOARP = 0x40  # A device which does not do neighbor discovery (ARP)
            NUD_PERMANENT = 0x80  # A static entry
            NUD_NONE = 0x00

        class Flag(object):
            NTF_USE = 0x01
            NTF_PROXY = 0x08  # A proxy ARP entry
            NTF_ROUTER = 0x80  # An IPv6 router

        class Attribute(object):
            # XXX : Not sure - starts at zero or one ... ??
            NDA_UNSPEC = 0x00  # Unknown type
            NDA_DST = 0x01  # A neighbour cache network. layer destination address
            NDA_LLADDR = 0x02  # A neighbor cache link layer address.
            NDA_CACHEINFO = 0x03  # Cache statistics
            NDA_PROBES = 0x04

    @classmethod
    def getNeighbors(cls):
        return cls.extract(Neighbor.Command.RTM_GETNEIGH)



# encoding: utf-8
"""
community.py

Created by Thomas Mangin on 2009-11-05.
Copyright (c) 2009-2017 Exa Networks. All rights reserved.
License: 3-clause BSD. (See the COPYRIGHT file)
"""

from struct import pack
from struct import unpack


# ==================================================================== Community
#


class Community(object):
    MAX = 0xFFFFFFFF

    NO_EXPORT = pack('!L', 0xFFFFFF01)
    NO_ADVERTISE = pack('!L', 0xFFFFFF02)
    NO_EXPORT_SUBCONFED = pack('!L', 0xFFFFFF03)
    NO_PEER = pack('!L', 0xFFFFFF04)
    BLACKHOLE = pack('!L', 0xFFFF029A)

    cache = {}
    caching = True

    __slots__ = ['community', '_str']

    def __init__(self, community):
        self.community = community
        if community == self.NO_EXPORT:
            self._str = 'no-export'
        elif community == self.NO_ADVERTISE:
            self._str = 'no-advertise'
        elif community == self.NO_EXPORT_SUBCONFED:
            self._str = 'no-export-subconfed'
        elif community == self.NO_PEER:
            self._str = 'no-peer'
        elif community == self.BLACKHOLE:
            self._str = 'blackhole'
        else:
            self._str = "%d:%d" % unpack('!HH', self.community)

    def __eq__(self, other):
        return self.community == other.community

    def __ne__(self, other):
        return self.community != other.community

    def __lt__(self, other):
        return self.community < other.community

    def __le__(self, other):
        return self.community <= other.community

    def __gt__(self, other):
        return self.community > other.community

    def __ge__(self, other):
        return self.community >= other.community

    def json(self):
        return "[ %d, %d ]" % unpack('!HH', self.community)

    def pack(self, negotiated=None):
        return self.community

    def __repr__(self):
        return self._str

    def __len__(self):
        return 4

    @classmethod
    def unpack(cls, community, direction, negotiated):
        return cls(community)

    @classmethod
    def cached(cls, community):
        if not cls.caching:
            return cls(community)
        if community in cls.cache:
            return cls.cache[community]
        instance = cls(community)
        cls.cache[community] = instance
        return instance


# Always cache well-known communities, they will be used a lot
if not Community.cache:
    Community.cache[Community.NO_EXPORT] = Community(Community.NO_EXPORT)
    Community.cache[Community.NO_ADVERTISE] = Community(Community.NO_ADVERTISE)
    Community.cache[Community.NO_EXPORT_SUBCONFED] = Community(Community.NO_EXPORT_SUBCONFED)
    Community.cache[Community.NO_PEER] = Community(Community.NO_PEER)
    Community.cache[Community.BLACKHOLE] = Community(Community.BLACKHOLE)



# encoding: utf-8
"""
rterid.py

Created by Evelio Vila on 2016-12-01.
Copyright (c) 2014-2017 Exa Networks. All rights reserved.
"""

from exabgp.protocol.ip import IP
from exabgp.bgp.message.notification import Notify

from exabgp.bgp.message.update.attribute.bgpls.linkstate import LINKSTATE


#   |    1030   | IPv4 Router-ID of   |   134/---    | [RFC5305]/4.3    |
#   |           | Remote Node         |              |                  |
#   |    1031   | IPv6 Router-ID of   |   140/---    | [RFC6119]/4.1    |
#   |           | Remote Node         |              |                  |


@LINKSTATE.register(lsid=1030)
@LINKSTATE.register(lsid=1031)
class RemoteTeRid(object):
    def __init__(self, terid):
        self.terid = terid

    def __repr__(self):
        return "Remote TE Router ID: %s" % (self.terid)

    @classmethod
    def unpack(cls, data, length):
        if len(data) == 4:
            # IPv4 address
            terid = IP.unpack(data[:4])
        elif len(data) == 16:
            # IPv6
            terid = IP.unpack(data[:16])
        return cls(terid=terid)

    def json(self, compact=None):
        return '"remote-te-router-id": "%s"' % str(self.terid)



# encoding: utf-8
"""
graceful.py

Created by Thomas Mangin on 2012-07-17.
Copyright (c) 2009-2017 Exa Networks. All rights reserved.
License: 3-clause BSD. (See the COPYRIGHT file)
"""

from struct import pack
from struct import unpack

from exabgp.vendoring import six

from exabgp.util import character
from exabgp.util import ordinal
from exabgp.util import concat_bytes_i
from exabgp.util import concat_bytes

from exabgp.protocol.family import AFI
from exabgp.protocol.family import SAFI
from exabgp.bgp.message.open.capability.capability import Capability

# =========================================================== Graceful (Restart)
# RFC 4727 - https://tools.ietf.org/html/rfc4727


@Capability.register()
class Graceful(Capability, dict):
    MAX = 0xFFFF
    ID = Capability.CODE.GRACEFUL_RESTART

    TIME_MASK = 0x0FFF
    FLAG_MASK = 0xF000

    # 0x8 is binary 1000
    RESTART_STATE = 0x08
    FORWARDING_STATE = 0x80

    def set(self, restart_flag, restart_time, protos):
        self.restart_flag = restart_flag
        self.restart_time = restart_time & Graceful.TIME_MASK
        for afi, safi, family_flag in protos:
            self[(afi, safi)] = family_flag & Graceful.FORWARDING_STATE
        return self

    def extract(self):
        restart = pack('!H', ((self.restart_flag << 12) | (self.restart_time & Graceful.TIME_MASK)))
        families = [(afi.pack(), safi.pack(), character(self[(afi, safi)])) for (afi, safi) in self.keys()]
        sfamilies = concat_bytes_i(concat_bytes(pafi, psafi, family) for (pafi, psafi, family) in families)
        return [concat_bytes(restart, sfamilies)]

    def __str__(self):
        families = [(str(afi), str(safi), hex(self[(afi, safi)])) for (afi, safi) in self.keys()]
        sfamilies = ' '.join(["%s/%s=%s" % (afi, safi, family) for (afi, safi, family) in families])
        return "Graceful Restart Flags %s Time %d %s" % (hex(self.restart_flag), self.restart_time, sfamilies)

    def json(self):
        d = {
            'name': '"graceful restart"',
            'time': self.restart_time,
            'address-family-flags': '{ %s}'
            % (
                ', '.join(
                    '"%s/%s": [%s ] ' % ('%s' % afi, safi, ' "restart"' if family & 0x80 else '')
                    for afi, safi, family in [(str(a), str(s), self[(a, s)]) for (a, s) in self.keys()]
                )
            ),
            'restart-flags': '[%s] ' % (' "forwarding" ' if self.restart_flag & 0x8 else ' '),
        }

        return '{ %s}' % ', '.join('"%s": %s' % (k, v) for k, v in six.iteritems(d))

    def families(self):
        return self.keys()

    @staticmethod
    def unpack_capability(instance, data, capability=None):  # pylint: disable=W0613
        # XXX: FIXME: should raise if instance was already setup
        restart = unpack('!H', data[:2])[0]
        restart_flag = restart >> 12
        restart_time = restart & Graceful.TIME_MASK
        data = data[2:]
        families = []
        while data:
            afi = AFI.unpack(data[:2])
            safi = SAFI.unpack(data[2])
            flag_family = ordinal(data[3])
            families.append((afi, safi, flag_family))
            data = data[4:]
        return instance.set(restart_flag, restart_time, families)



# encoding: utf-8
"""
match.py

Created by Thomas Mangin on 2015-06-22.
Copyright (c) 2009-2017 Exa Networks. All rights reserved.
License: 3-clause BSD. (See the COPYRIGHT file)
"""

from exabgp.configuration.core import Section

from exabgp.configuration.flow.parser import source
from exabgp.configuration.flow.parser import destination
from exabgp.configuration.flow.parser import any_port
from exabgp.configuration.flow.parser import source_port
from exabgp.configuration.flow.parser import destination_port
from exabgp.configuration.flow.parser import tcp_flags
from exabgp.configuration.flow.parser import protocol
from exabgp.configuration.flow.parser import next_header
from exabgp.configuration.flow.parser import fragment
from exabgp.configuration.flow.parser import packet_length
from exabgp.configuration.flow.parser import icmp_code
from exabgp.configuration.flow.parser import icmp_type
from exabgp.configuration.flow.parser import dscp
from exabgp.configuration.flow.parser import traffic_class
from exabgp.configuration.flow.parser import flow_label


class ParseFlowMatch(Section):
    definition = [
        'source 10.0.0.0/24',
        'source ::1/128/0',
        'destination 10.0.1.0/24',
        'port 25',
        'source-port >1024',
        'destination-port [ =80 =3128 >8080&<8088 ]',
        'packet-length [ >200&<300 >400&<500 ]',
        'tcp-flags [ 0x20+0x8+0x1 #name-here ]  # to check',
        '(ipv4 only) protocol [ udp tcp ]',
        '(ipv4 only) fragment [ dont-fragment is-fragment first-fragment last-fragment ]',
        '(ipv6 only) next-header [ udp tcp ]',
        '(ipv6 only) flow-label >100&<2000',
        '(ipv6 only) icmp-type 35  # to check',
        '(ipv6 only) icmp-code 55  # to check',
    ]

    syntax = 'match {\n' '  %s;\n' '}' % ';\n  '.join(definition)

    known = {
        'source': source,
        'source-ipv4': source,
        'source-ipv6': source,
        'destination': destination,
        'destination-ipv4': destination,
        'destination-ipv6': destination,
        'protocol': protocol,
        'next-header': next_header,
        'port': any_port,
        'destination-port': destination_port,
        'source-port': source_port,
        'icmp-type': icmp_type,
        'icmp-code': icmp_code,
        'tcp-flags': tcp_flags,
        'packet-length': packet_length,
        'dscp': dscp,
        'traffic-class': traffic_class,
        'fragment': fragment,
        'flow-label': flow_label,
    }

    # 'source-ipv4','destination-ipv4',

    action = {
        'source': 'nlri-add',
        'source-ipv4': 'nlri-add',
        'source-ipv6': 'nlri-add',
        'destination': 'nlri-add',
        'destination-ipv4': 'nlri-add',
        'destination-ipv6': 'nlri-add',
        'port': 'nlri-add',
        'source-port': 'nlri-add',
        'destination-port': 'nlri-add',
        'protocol': 'nlri-add',
        'packet-length': 'nlri-add',
        'tcp-flags': 'nlri-add',
        'next-header': 'nlri-add',
        'fragment': 'nlri-add',
        'icmp-code': 'nlri-add',
        'icmp-type': 'nlri-add',
        'packet-length': 'nlri-add',
        'dscp': 'nlri-add',
        'traffic-class': 'nlri-add',
        'flow-label': 'nlri-add',
    }

    name = 'flow/match'

    def __init__(self, tokeniser, scope, error, logger):
        Section.__init__(self, tokeniser, scope, error, logger)

    def clear(self):
        pass

    def pre(self):
        return True

    def post(self):
        return True
